import pandas as pd
import numpy as np
from typing import Callable, List, Optional, Tuple

from sklearn.base import BaseEstimator
from qolmat.imputations.imputers import ImputerRegressor
from qolmat.utils.exceptions import PyTorchExtraNotInstalled

try:
    import torch
    import torch.nn as nn
    import torch.optim as optim
except ModuleNotFoundError:
    raise PyTorchExtraNotInstalled


def build_mlp_example(
    input_dim: int,
    list_num_neurons: List[int],
    output_dim: int = 1,
    activation: Callable = nn.ReLU,
) -> nn.Sequential:
    """
    Constructs a multi-layer perceptron (MLP) with a custom architecture.

    Parameters
    ----------
    input_dim : int
        Dimension of the input layer.
    list_num_neurons : List[int]
        List specifying the number of neurons in each hidden layer.
    output_dim : int, optional
        Dimension of the output layer, defaults to 1.
    activation : nn.Module, optional
        Activation function to use between hidden layers, defaults to nn.ReLU().

    Returns
    -------
    nn.Sequential
        PyTorch model representing the MLP.

    Raises
    ------
    TypeError
        If `input_dim` is not an integer or `list_num_neurons` is not a list.

    Examples
    --------
    >>> model = build_mlp_example(input_dim=10, list_num_neurons=[32, 64, 128], output_dim=1)
    >>> print(model)
    Sequential(
      (0): Linear(in_features=10, out_features=32, bias=True)
      (1): ReLU()
      (2): Linear(in_features=32, out_features=64, bias=True)
      (3): ReLU()
      (4): Linear(in_features=64, out_features=128, bias=True)
      (5): ReLU()
      (6): Linear(in_features=128, out_features=1, bias=True)
    )
    """
    layers = []
    for num_neurons in list_num_neurons:
        layers.append(nn.Linear(input_dim, num_neurons))
        layers.append(activation())
        input_dim = num_neurons
    layers.append(nn.Linear(input_dim, output_dim))

    estimator = nn.Sequential(*layers)
    return estimator


class ImputerRegressorPyTorch(ImputerRegressor):
    def __init__(
        self,
        groups: Tuple[str, ...] = (),
        estimator: Optional[BaseEstimator] = None,
        handler_nan: str = "column",
        epochs: int = 100,
        learning_rate: float = 0.01,
        loss_fn: Callable = nn.L1Loss(),
    ):
        super().__init__(
            imputer_params=("handler_nan", "epochs", "monitor", "patience"),
            groups=groups,
            handler_nan=handler_nan,
        )
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.loss_fn = loss_fn
        self.estimator = estimator

    def _fit_estimator(self, X: pd.DataFrame, y: pd.DataFrame, **hp):
        """
        Fit the PyTorch estimator using the provided input and target data.

        Parameters
        ----------
        X : pd.DataFrame
            The input data for training.
        y : pd.DataFrame
            The target data for training.
        """
        optimizer = optim.Adam(self.estimator.parameters(), lr=self.learning_rate)
        loss_fn = self.loss_fn
        if self.estimator is None:
            assert AssertionError("Estimator is not provided.")
        else:
            for epoch in range(self.epochs):
                self.estimator.train()
                optimizer.zero_grad()

                input_data = torch.Tensor(X.values)
                target_data = torch.Tensor(y.values)
                target_data = target_data.unsqueeze(1)
                outputs = self.estimator(input_data)
                loss = loss_fn(outputs, target_data)

                loss.backward()
                optimizer.step()
                if (epoch + 1) % 10 == 0:
                    print(f"Epoch [{epoch+1}/{self.epochs}], Loss: {loss.item():.4f}")

    def _predict_estimator(self, X: pd.DataFrame) -> pd.Series:
        """
        Perform predictions using the trained PyTorch estimator.

        Parameters
        ----------
        X : pd.DataFrame
            The input data for prediction.

        Returns
        -------
        pd.Series
            The predicted values.
        """
        if self.estimator:
            input_data = torch.Tensor(X.values)
            output_data = self.estimator(input_data)
            pred = output_data.detach().numpy().flatten()
            return pd.Series(pred, index=X.index, dtype=float)
        assert AssertionError("Estimator is not provided.")



import numpy as np
import pandas as pd
import pytest

from qolmat.benchmark import missing_patterns as mp

df_complet = pd.DataFrame({"col1": [i for i in range(100)], "col2": [2 * i for i in range(100)]})

df_incomplet = df_complet.copy()
df_incomplet.iloc[99, :] = np.nan

index_group = np.concatenate([[f"g{i}" for i in range(10)] for j in range(10)])

df_complet_group = df_complet.copy()
df_complet_group = df_complet_group.set_index(index_group)
df_complet_group.index = df_complet_group.index.set_names("group")

df_incomplet_group = df_incomplet.copy()
df_incomplet_group = df_incomplet_group.set_index(pd.Index(index_group))
df_incomplet_group.index = df_incomplet_group.index.set_names("group")

list_generators = {
    "geo": mp.GeometricHoleGenerator(n_splits=2, ratio_masked=0.1, random_state=42),
    "unif": mp.UniformHoleGenerator(n_splits=2, ratio_masked=0.1, random_state=42),
    "multi": mp.MultiMarkovHoleGenerator(n_splits=2, ratio_masked=0.1, random_state=42),
    "group": mp.GroupedHoleGenerator(
        n_splits=2, ratio_masked=0.1, random_state=42, groups=("group",)
    ),
}


@pytest.mark.parametrize(
    "df, generator",
    [
        (df_incomplet, list_generators["geo"]),
        (df_incomplet, list_generators["unif"]),
        (df_incomplet, list_generators["multi"]),
        (df_incomplet_group, list_generators["group"]),
    ],
)
def test_SamplerHoleGenerator_split(df: pd.DataFrame, generator: mp._HoleGenerator) -> None:
    mask = generator.split(df)[0]
    col1_holes = mask["col1"].sum()
    col2_holes = mask["col2"].sum()
    expected_col1_holes = 10
    expected_col2_holes = 10
    np.testing.assert_allclose(col1_holes, expected_col1_holes, atol=1)
    np.testing.assert_allclose(col2_holes, expected_col2_holes, atol=1)


@pytest.mark.parametrize(
    "df, generator",
    [
        (df_complet, list_generators["geo"]),
        (df_complet, list_generators["unif"]),
        (df_complet, list_generators["multi"]),
        (df_complet_group, list_generators["group"]),
    ],
)
def test_SamplerHoleGenerator_without_real_nans(
    df: pd.DataFrame, generator: mp._HoleGenerator
) -> None:
    real_nan = np.random.choice([True, False], size=df.size, p=[0.4, 0.6]).reshape(100, 2)
    df[real_nan] = np.nan

    mask = generator.split(df)[0]

    loc_real_nans_col1 = np.where(df["col1"].isna())[0]
    loc_mask_col1 = np.where(mask["col1"])[0]

    loc_real_nans_col2 = np.where(df["col2"].isna())[0]
    loc_mask_col2 = np.where(mask["col2"])[0]

    np.testing.assert_allclose(len(set(loc_real_nans_col1) & set(loc_mask_col1)), 0)
    np.testing.assert_allclose(len(set(loc_real_nans_col2) & set(loc_mask_col2)), 0)



import pytest
from qolmat.utils import exceptions


def test_utils_exception_init():
    try:
        raise exceptions.PyTorchExtraNotInstalled()
    except exceptions.PyTorchExtraNotInstalled as e:
        assert (
            str(e)
            == """Please install torch xx.xx.xx
        pip install qolmat[pytorch]"""
        )



from functools import partial
from typing import Callable, Dict, List, Optional

import numpy as np
import pandas as pd
import scipy
from scipy.stats import wasserstein_distance as dist_wasserstein_1d
import sklearn
from sklearn import metrics as skm
from sklearn.ensemble import BaseEnsemble
from sklearn.preprocessing import StandardScaler
import dcor

from qolmat.utils.exceptions import NotEnoughSamples

EPS = np.finfo(float).eps

###########################
# Column-wise metrics     #
###########################


def columnwise_metric(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame, metric: Callable, **kwargs
) -> pd.Series:
    """For each column, compute a metric score based on the true dataframe
    and the predicted dataframe

    Parameters
    ----------
    df1 : pd.DataFrame
        True dataframe
    df2 : pd.DataFrame
        Predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on
    metric : Callable
        metric function

    Returns
    -------
    pd.Series
        Series of scores for all columns
    """
    values = {}
    for col in df1.columns:
        df1_col = df1.loc[df_mask[col], col]
        df2_col = df2.loc[df_mask[col], col]
        assert df1_col.notna().all()
        assert df2_col.notna().all()
        values[col] = metric(df1_col, df2_col, **kwargs)
    return pd.Series(values)


def mean_squared_error(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame) -> pd.Series:
    """Mean squared error between two dataframes.

    Parameters
    ----------
    df1 : pd.DataFrame
        True dataframe
    df2 : pd.DataFrame
        Predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on

    Returns
    -------
    pd.Series
    """
    return columnwise_metric(df1, df2, df_mask, skm.mean_squared_error)


def root_mean_squared_error(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> pd.Series:
    """Root mean squared error between two dataframes.

    Parameters
    ----------
    df1 : pd.DataFrame
        True dataframe
    df2 : pd.DataFrame
        Predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on

    Returns
    -------
    pd.Series
    """
    return columnwise_metric(df1, df2, df_mask, skm.mean_squared_error, squared=False)


def mean_absolute_error(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame) -> pd.Series:
    """Mean absolute error between two dataframes.

    Parameters
    ----------
    df1 : pd.DataFrame
        True dataframe
    df2 : pd.DataFrame
        Predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on

    Returns
    -------
    pd.Series
    """
    return columnwise_metric(df1, df2, df_mask, skm.mean_absolute_error)


def mean_absolute_percentage_error(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> pd.Series:
    """Mean absolute percentage error between two dataframes.

    Parameters
    ----------
    df1 : pd.DataFrame
        True dataframe
    df2 : pd.DataFrame
        Predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on

    Returns
    -------
    pd.Series
    """
    return columnwise_metric(df1, df2, df_mask, skm.mean_absolute_percentage_error)


def _weighted_mean_absolute_percentage_error_1D(values1: pd.Series, values2: pd.Series) -> float:
    """Weighted mean absolute percentage error between two series.
    Based on https://en.wikipedia.org/wiki/Mean_absolute_percentage_error

    Parameters
    ----------
    values1 : pd.Series
        true series
    values2 : pd.Series
        predicted series

    Returns
    -------
    float
        Weighted mean absolute percentage error
    """
    return (values1 - values2).abs().sum() / values1.abs().sum()


def weighted_mean_absolute_percentage_error(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> pd.Series:
    """Weighted mean absolute percentage error between two dataframes.

    Parameters
    ----------
    df1 : pd.DataFrame
        True dataframe
    df2 : pd.DataFrame
        Predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on

    Returns
    -------
    pd.Series
    """
    return columnwise_metric(df1, df2, df_mask, _weighted_mean_absolute_percentage_error_1D)


def dist_wasserstein(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame, method: str = "columnwise"
) -> pd.Series:
    """Wasserstein distances between columns of 2 dataframes.
    Wasserstein distance can only be computed columnwise

    Parameters
    ----------
    df1 : pd.DataFrame
        True dataframe
    df2 : pd.DataFrame
        Predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on

    Returns
    -------
    pd.Series
        wasserstein distances
    """
    if method == "columnwise":
        return columnwise_metric(df1, df2, df_mask, scipy.stats.wasserstein_distance)
    else:
        raise AssertionError(
            f"The parameter of the function wasserstein_distance should be one of"
            f"the following: [`columnwise`], not `{method}`!"
        )


def _get_numerical_features(df1: pd.DataFrame) -> List[str]:
    """Get numerical features from dataframe

    Parameters
    ----------
    df1 : pd.DataFrame

    Returns
    -------
    List[str]
        List of numerical features

    Raises
    ------
    Exception
        No numerical feature is found
    """
    cols_numerical = df1.select_dtypes(include=np.number).columns.tolist()
    if len(cols_numerical) == 0:
        raise Exception("No numerical feature is found.")
    else:
        return cols_numerical


def _get_categorical_features(df1: pd.DataFrame) -> List[str]:
    """Get categorical features from dataframe

    Parameters
    ----------
    df1 : pd.DataFrame

    Returns
    -------
    List[str]
        List of categorical features

    Raises
    ------
    Exception
        No categorical feature is found
    """

    cols_numerical = df1.select_dtypes(include=np.number).columns.tolist()
    cols_categorical = [col for col in df1.columns.to_list() if col not in cols_numerical]
    if len(cols_categorical) == 0:
        raise Exception("No categorical feature is found.")
    else:
        return cols_categorical


def kolmogorov_smirnov_test_1D(df1: pd.Series, df2: pd.Series) -> float:
    """Compute KS test statistic of the two-sample Kolmogorov-Smirnov test for goodness of fit.
    See more in https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.ks_2samp.html.

    Parameters
    ----------
    df1 : pd.Series
        true series
    df2 : pd.Series
        predicted series

    Returns
    -------
    float
        KS test statistic
    """
    return scipy.stats.ks_2samp(df1, df2)[0]


def kolmogorov_smirnov_test(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> pd.Series:
    """Kolmogorov Smirnov Test for numerical features.
    Lower score means better performance.

    Parameters
    ----------
    df1 : pd.DataFrame
        true dataframe
    df2 : pd.DataFrame
        predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on

    Returns
    -------
    pd.Series
        KS test statistic
    """
    cols_numerical = _get_numerical_features(df1)
    return columnwise_metric(
        df1[cols_numerical],
        df2[cols_numerical],
        df_mask[cols_numerical],
        kolmogorov_smirnov_test_1D,
    )


def _total_variance_distance_1D(df1: pd.Series, df2: pd.Series) -> float:
    """Compute Total Variance Distance for a categorical feature
    It is based on TVComplement in https://github.com/sdv-dev/SDMetrics

    Parameters
    ----------
    df1 : pd.Series
        true series
    df2 : pd.Series
        predicted series

    Returns
    -------
    float
        Total variance distance
    """
    list_categories = list(set(df1.unique()).union(set(df2.unique())))
    freqs1 = df1.value_counts() / len(df1)
    freqs1 = freqs1.reindex(list_categories, fill_value=0.0)
    freqs2 = df2.value_counts() / len(df2)
    freqs2 = freqs2.reindex(list_categories, fill_value=0.0)
    return (freqs1 - freqs2).abs().sum()


def total_variance_distance(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> pd.Series:
    """Total variance distance for categorical features
    It is based on TVComplement in https://github.com/sdv-dev/SDMetrics

    Parameters
    ----------
    df1 : pd.DataFrame
        true dataframe
    df2 : pd.DataFrame
        predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on

    Returns
    -------
    pd.Series
        Total variance distance
    """
    cols_categorical = _get_categorical_features(df1)
    return columnwise_metric(
        df1[cols_categorical],
        df2[cols_categorical],
        df_mask[cols_categorical],
        _total_variance_distance_1D,
    )


def _check_same_number_columns(df1: pd.DataFrame, df2: pd.DataFrame):
    if len(df1.columns) != len(df2.columns):
        raise Exception("inputs have to have the same number of columns.")


def _get_correlation_pearson_matrix(df: pd.DataFrame, use_p_value: bool = True) -> pd.DataFrame:
    """Get matrix of correlation values for numerical features
    based on Pearson correlation coefficient or p-value for testing non-correlation.

    Parameters
    ----------
    df : pd.DataFrame
        dataframe
    use_p_value : bool, optional
        use the p-value instead of the correlation coefficient, by default True

    Returns
    -------
    pd.DataFrame
        Correlation matrix
    """
    cols = df.columns.tolist()
    matrix = np.zeros((len(df.columns), len(df.columns)))
    for idx_1, col_1 in enumerate(cols):
        for idx_2, col_2 in enumerate(cols):
            res = scipy.stats.mstats.pearsonr(
                df[col_1].array.reshape(-1, 1), df[col_2].array.reshape(-1, 1)
            )
            if use_p_value:
                matrix[idx_1, idx_2] = res[1]
            else:
                matrix[idx_1, idx_2] = res[0]

    return pd.DataFrame(matrix, index=cols, columns=cols)


def mean_difference_correlation_matrix_numerical_features(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    df_mask: pd.DataFrame,
    use_p_value: bool = True,
) -> pd.Series:
    """Mean absolute of differences between the correlation matrices of df1 and df2.
    based on Pearson correlation coefficient or p-value for testing non-correlation.

    Parameters
    ----------
    df1 : pd.DataFrame
        true dataframe
    df2 : pd.DataFrame
        predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on
    use_p_value : bool, optional
        use the p-value instead of the correlation coefficient, by default True

    Returns
    -------
    pd.Series
        Mean absolute of differences for each feature
    """
    df1 = df1[df_mask].dropna(axis=0)
    df2 = df2[df_mask].dropna(axis=0)

    _check_same_number_columns(df1, df2)

    cols_numerical = _get_numerical_features(df1)
    df_corr1 = _get_correlation_pearson_matrix(df1[cols_numerical], use_p_value=use_p_value)
    df_corr2 = _get_correlation_pearson_matrix(df2[cols_numerical], use_p_value=use_p_value)

    diff_corr = (df_corr1 - df_corr2).abs().mean(axis=1)
    return pd.Series(diff_corr, index=cols_numerical)


def _get_correlation_chi2_matrix(data: pd.DataFrame, use_p_value: bool = True) -> pd.DataFrame:
    """Get matrix of correlation values for categorical features
    based on Chi-square test of independence of variables (the test statistic or the p-value).

    Parameters
    ----------
    df : pd.DataFrame
        dataframe
    use_p_value : bool, optional
        use the p-value of the test instead of the test statistic, by default True

    Returns
    -------
    pd.DataFrame
        Correlation matrix
    """
    cols = data.columns.tolist()
    matrix = np.zeros((len(data.columns), len(data.columns)))
    for idx_1, col_1 in enumerate(cols):
        for idx_2, col_2 in enumerate(cols):
            freq = data.pivot_table(
                index=col_1, columns=col_2, aggfunc="size", fill_value=0
            ).to_numpy()
            res = scipy.stats.chi2_contingency(freq)
            if use_p_value:
                matrix[idx_1, idx_2] = res[1]
            else:
                matrix[idx_1, idx_2] = res[0]
    return pd.DataFrame(matrix, index=cols, columns=cols)


def mean_difference_correlation_matrix_categorical_features(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    df_mask: pd.DataFrame,
    use_p_value: bool = True,
) -> pd.Series:
    """Mean absolute of differences between the correlation matrix of df1 and df2
    based on Chi-square test of independence of variables (the test statistic or the p-value)

    Parameters
    ----------
    df1 : pd.DataFrame
        true dataframe
    df2 : pd.DataFrame
        predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on
    use_p_value : bool, optional
        use the p-value of the test instead of the test statistic, by default True

    Returns
    -------
    pd.Series
        Mean absolute of differences for each feature
    """
    df1 = df1[df_mask].dropna(axis=0)
    df2 = df2[df_mask].dropna(axis=0)

    _check_same_number_columns(df1, df2)

    cols_categorical = _get_categorical_features(df1)
    df_corr1 = _get_correlation_chi2_matrix(df1[cols_categorical], use_p_value=use_p_value)
    df_corr2 = _get_correlation_chi2_matrix(df2[cols_categorical], use_p_value=use_p_value)

    diff_corr = (df_corr1 - df_corr2).abs().mean(axis=1)
    return pd.Series(diff_corr, index=cols_categorical)


def _get_correlation_f_oneway_matrix(
    df: pd.DataFrame,
    cols_categorical: List[str],
    cols_numerical: List[str],
    use_p_value: bool = True,
) -> pd.DataFrame:
    """Get matrix of correlation values between categorical and numerical features
    based on the one-way ANOVA.

    Parameters
    ----------
    df : pd.DataFrame
        dataframe
    cols_categorical : List[str]
        list categorical columns
    cols_numerical : List[str]
        list numerical columns
    use_p_value : bool, optional
        use the p-value of the test instead of the test statistic, by default True

    Returns
    -------
    pd.DataFrame
        Correlation matrix
    """
    matrix = np.zeros((len(cols_categorical), len(cols_numerical)))
    for idx_cat, col_cat in enumerate(cols_categorical):
        for idx_num, col_num in enumerate(cols_numerical):
            category_group_lists = df.groupby(col_cat)[col_num].apply(list)
            res = scipy.stats.f_oneway(*category_group_lists)
            if use_p_value:
                matrix[idx_cat, idx_num] = res[1]
            else:
                matrix[idx_cat, idx_num] = res[0]
    return pd.DataFrame(matrix, index=cols_categorical, columns=cols_numerical)


def mean_diff_corr_matrix_categorical_vs_numerical_features(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    df_mask: pd.DataFrame,
    use_p_value: bool = True,
) -> pd.Series:
    """Mean absolute of differences between the correlation matrix of df1 and df2
    based on the one-way ANOVA.

    Parameters
    ----------
    df1 : pd.DataFrame
        true dataframe
    df2 : pd.DataFrame
        predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on
    use_p_value : bool, optional
        use the p-value of the test instead of the test statistic, by default True

    Returns
    -------
    pd.Series
        Mean absolute of differences for each feature
    """
    df1 = df1[df_mask].dropna(axis=0)
    df2 = df2[df_mask].dropna(axis=0)

    _check_same_number_columns(df1, df2)

    cols_categorical = _get_categorical_features(df1)
    cols_numerical = _get_numerical_features(df1)
    df_corr1 = _get_correlation_f_oneway_matrix(
        df1, cols_categorical, cols_numerical, use_p_value=use_p_value
    )
    df_corr2 = _get_correlation_f_oneway_matrix(
        df2, cols_categorical, cols_numerical, use_p_value=use_p_value
    )
    diff_corr = (df_corr1 - df_corr2).abs().mean(axis=1)
    return pd.Series(diff_corr, index=cols_categorical)


###########################
# Row-wise metrics        #
###########################


def _sum_manhattan_distances_1D(values: pd.Series) -> float:
    """Sum of Manhattan distances computed for one column
    It is based on https://www.geeksforgeeks.org/sum-manhattan-distances-pairs-points/

    Parameters
    ----------
    values : pd.Series
        Values of a column

    Returns
    -------
    float
        Sum of Manhattan distances
    """
    values = values.sort_values(ascending=True)
    sums_partial = values.shift().fillna(0.0).cumsum()
    differences_partial = values * np.arange(len(values)) - sums_partial
    res = differences_partial.sum()
    return res


def _sum_manhattan_distances(df1: pd.DataFrame) -> float:
    """Sum Manhattan distances between all pairs of rows.
    It is based on https://www.geeksforgeeks.org/sum-manhattan-distances-pairs-points/

    Parameters
    ----------
    df1 : pd.DataFrame

    Returns
    -------
    float
        Sum of Manhattan distances for all pairs of rows.
    """
    cols = df1.columns.tolist()
    result = sum([_sum_manhattan_distances_1D(df1[col]) for col in cols])
    return result


def sum_energy_distances(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame) -> pd.Series:
    """Sum of energy distances between df1 and df2.
    It is based on https://dcor.readthedocs.io/en/latest/theory.html#

    Parameters
    ----------
    df1 : pd.DataFrame
        true dataframe
    df2 : pd.DataFrame
        predicted dataframe
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on

    Returns
    -------
    pd.Series
        Sum of energy distances between df1 and df2.
    """

    # Replace nan in dataframe
    df1 = df1[df_mask].fillna(0.0)
    df2 = df2[df_mask].fillna(0.0)

    # sum of (len_df1 * (len_df1 - 1) / 2) distances for df1
    sum_distances_df1 = _sum_manhattan_distances(df1)
    sum_distances_df2 = _sum_manhattan_distances(df2)

    df = pd.concat([df1, df2])
    sum_distances_df1_df2 = _sum_manhattan_distances(df)
    sum_distance = 2 * sum_distances_df1_df2 - 4 * sum_distances_df1 - 4 * sum_distances_df2

    return pd.Series(sum_distance, index=["All"])


def sum_pairwise_distances(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame, metric: str = "cityblock"
) -> float:
    """Sum of pairwise distances based on a predefined metric.
    Metrics are found in this link
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.cdist.html

    Parameters
    ----------
    df1 : pd.DataFrame
        First empirical distribution without nans
    df2 : pd.DataFrame
        Second empirical distribution without nans
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on
    metric : str, optional
        distance metric, by default 'cityblock'

    Returns
    -------
    float
        Sum of pairwise distances based on a predefined metric
    """
    df1 = df1[df_mask.any(axis=1)]
    df2 = df2[df_mask.any(axis=1)]
    distances = np.sum(scipy.spatial.distance.cdist(df1, df2, metric=metric))

    return distances


###########################
# Dataframe-wise metrics  #
###########################


def frechet_distance(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    df_mask: pd.DataFrame,
) -> float:
    """Compute the Fréchet distance between two dataframes df1 and df2
    frechet_distance = || mu_1 - mu_2 ||_2^2 + Tr(Sigma_1 + Sigma_2 - 2(Sigma_1 . Sigma_2)^(1/2))
    It is normalized, df1 and df_ are first scaled by a factor
        (std(df1) + std(df2)) / 2
    and then centered around
        (mean(df1) + mean(df2)) / 2

    Dowson, D. C., and BV666017 Landau. "The Fréchet distance between multivariate normal
    distributions."
    Journal of multivariate analysis 12.3 (1982): 450-455.

    Parameters
    ----------
    df1 : pd.DataFrame
        true dataframe
    df2 : pd.DataFrame
        predicted dataframe
    df_mask : pd.DataFrame
        Mask indicating on which values the distance has to computed on

    Returns
    -------
    float
        frechet_distance
    """

    if df1.shape != df2.shape:
        raise Exception("inputs have to be of same dimensions.")

    df_true = df1[df_mask.any(axis=1)]
    df_pred = df2[df_mask.any(axis=1)]

    std = (np.std(df_true) + np.std(df_pred) + EPS) / 2
    mu = (np.nanmean(df_true, axis=0) + np.nanmean(df_pred, axis=0)) / 2
    df_true = (df_true - mu) / std
    df_pred = (df_pred - mu) / std

    mu_true = np.nanmean(df_true, axis=0)
    sigma_true = np.ma.cov(np.ma.masked_invalid(df_true), rowvar=False).data
    mu_pred = np.nanmean(df_pred, axis=0)
    sigma_pred = np.ma.cov(np.ma.masked_invalid(df_pred), rowvar=False).data

    ssdiff = np.sum((mu_true - mu_pred) ** 2.0)
    product = np.array(sigma_true @ sigma_pred)
    if product.ndim < 2:
        product = product.reshape(-1, 1)
    covmean = scipy.linalg.sqrtm(product)
    if np.iscomplexobj(covmean):
        covmean = covmean.real
    frechet_dist = ssdiff + np.trace(sigma_true + sigma_pred - 2.0 * covmean)

    return frechet_dist / df_true.shape[0]


def frechet_distance_pattern(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    df_mask: pd.DataFrame,
) -> pd.Series:
    """Frechet distance computed using a pattern decomposition

    Parameters
    ----------
    df1 : pd.DataFrame
        First empirical ditribution
    df2 : pd.DataFrame
        Second empirical ditribution
    df_mask : pd.DataFrame
        Mask indicating on which values the distance has to computed on

    Returns
    -------
    pd.Series
        Series of computed metrics
    """
    cols_numerical = _get_numerical_features(df1)
    distance = frechet_distance(df1[cols_numerical], df2[cols_numerical], df_mask[cols_numerical])
    return pd.Series(distance, index=["All"])


def density_from_rf(
    df: pd.DataFrame, estimator: BaseEnsemble, df_est: Optional[pd.DataFrame] = None
):
    """Estimates the density of the empirical distribution given by df at the sample points given
    by df_est. The estimation uses an random forest estimator and relies on the average number of
    samples in the leaf corresponding to each estimation point.

    Disclaimer: this method is experimental and has no known theoretical grounds

    Parameters
    ----------
    df : pd.DataFrame
        Empirical distribution which density should be estimated
    estimator : BaseEnsemble
        Estimator defining the forest upon which is based the density counting.
    df_est : pd.DataFrame, optional
        Sample points of the estimation, by default None
        If None, the density is estimated at the points given by `df`.

    Returns
    -------
    pd.Series
        Series of floats providing the normalized density
    """
    if df_est is None:
        df_est = df.copy()
    if df_est.index.names == [None]:
        cols_index = ["index"]
    else:
        cols_index = df_est.index.names
    counts = pd.Series(0, index=df_est.index)
    df_leafs = pd.DataFrame(estimator.apply(df), index=df.index)
    df_leafs_est = pd.DataFrame(estimator.apply(df_est), index=df_est.index)
    for i_tree in range(estimator.n_estimators):
        leafs = df_leafs[i_tree].rename("id_leaf")
        leafs_est = df_leafs_est[i_tree].rename("id_leaf")
        counts_leafs = leafs.value_counts().rename("count")
        df_merge = pd.merge(leafs_est.reset_index(), counts_leafs.reset_index(), on="id_leaf")
        df_merge = df_merge.set_index(cols_index)
        counts += df_merge["count"]
    counts /= counts.sum()
    return counts


def kl_divergence_1D(df1: pd.Series, df2: pd.Series) -> float:
    """Estimation of the Kullback-Leibler divergence between the two 1D empirical distributions
    given by `df1`and `df2`. The samples are binarized using a uniform spacing with 20 bins from
    the smallest to the largest value. Not that this may be a coarse estimation.

    Parameters
    ----------
    df1 : pd.Series
        First empirical distribution
    df2 : pd.Series
        Second empirical distribution

    Returns
    -------
    float
        Kullback-Leibler divergence between the two empirical distributions.
    """
    min_val = min(df1.min(), df2.min())
    max_val = max(df1.max(), df2.max())
    bins = np.linspace(min_val, max_val, 20)
    p = np.histogram(df1, bins=bins, density=True)[0]
    q = np.histogram(df2, bins=bins, density=True)[0]
    return scipy.stats.entropy(p + EPS, q + EPS)


def kl_divergence_gaussian_exact(
    mean1: pd.Series, cov1: pd.DataFrame, mean2: pd.Series, cov2: pd.DataFrame
) -> float:
    """Exact Kullback-Leibler divergence computed between two multivariate normal distributions

    Parameters
    ----------
    mean1: pd.Series
        Mean of the first distribution
    cov1: pd.DataFrame
        Covariance matrx of the first distribution
    mean2: pd.Series
        Mean of the second distribution
    cov2: pd.DataFrame
        Covariance matrx of the second distribution
    Returns
    -------
    float
        Kulback-Leibler divergence
    """
    n_variables = len(mean1)
    L1, lower1 = scipy.linalg.cho_factor(cov1)
    L2, lower2 = scipy.linalg.cho_factor(cov2)
    M = scipy.linalg.solve(L2, L1)
    y = scipy.linalg.solve(L2, mean2 - mean1)
    norm_M = (M**2).sum().sum()
    norm_y = (y**2).sum()
    term_diag_L = 2 * np.sum(np.log(np.diagonal(L2) / np.diagonal(L1)))
    print(norm_M, "-", n_variables, "+", norm_y, "+", term_diag_L)
    div_kl = 0.5 * (norm_M - n_variables + norm_y + term_diag_L)
    return div_kl


def kl_divergence_gaussian(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.Series) -> float:
    """Kullback-Leibler divergence estimation based on a Gaussian approximation of both empirical
    distributions

    Parameters
    ----------
    df1 : pd.DataFrame
        First empirical distribution
    df2 : pd.DataFrame
        Second empirical distribution
    df_mask: pd.DataFrame
        Mask indicating on what values the divergence should be computed

    Returns
    -------
    pd.Series
        Series of estimated metrics
    """
    df1 = df1[df_mask.any(axis=1)]
    df2 = df2[df_mask.any(axis=1)]
    cov1 = df1.cov()
    cov2 = df2.cov()
    mean1 = df1.mean()
    mean2 = df2.mean()

    div_kl = kl_divergence_gaussian_exact(mean1, cov1, mean2, cov2)
    return div_kl


def kl_divergence_forest(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame) -> float:
    """Kullback-Leibler divergence estimation based on a random forest fitted on the first
    empirical distribution

    Disclaimer: this method is experimental and has no known theoretical grounds

    Parameters
    ----------
    df1 : pd.DataFrame
        First empirical distribution
    df2 : pd.DataFrame
        Second empirical distribution
    df_mask: pd.DataFrame
        Mask indicating on what values the divergence should be computed

    Returns
    -------
    pd.Series
        Series of estimated metrics
    """
    df1 = df1[df_mask.any(axis=1)]
    df2 = df2[df_mask.any(axis=1)]
    # df_1 = StandardScaler().fit_transform(df1[df_mask.any(axis=1)])
    # df_2 = StandardScaler().fit_transform(df2[df_mask.any(axis=1)])
    n_estimators = 100
    # estimator = sklearn.ensemble.RandomForestClassifier(
    #     n_estimators=n_estimators, max_depth=10
    # )
    # X = pd.concat([df1, df2])
    # y = pd.concat([pd.Series([False] * len(df1)), pd.Series([True] * len(df2))])
    # estimator.fit(X, y)
    estimator = sklearn.ensemble.RandomTreesEmbedding(n_estimators=n_estimators, random_state=123)
    estimator.fit(df1)
    counts1 = density_from_rf(df1, estimator, df_est=df2)
    counts2 = density_from_rf(df2, estimator, df_est=df2)
    div_kl = np.mean(np.log(counts1 / counts2) * counts1 / counts2)
    return div_kl


def kl_divergence(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    df_mask: pd.Series,
    method: str = "columnwise",
    min_n_rows: int = 10,
) -> pd.Series:
    """
    Estimation of the Kullback-Leibler divergence between too empirical distributions. Three
    methods are implemented:
    - columnwise, relying on a uniform binarization and only taking marginals into account
    (https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence),
    - gaussian, relying on a Gaussian approximation,
    - random_forest, experimental

    Parameters
    ----------
    df1 : pd.DataFrame
        First empirical distribution
    df2 : pd.DataFrame
        Second empirical distribution
    df_mask: pd.DataFrame
        Mask indicating on what values the divergence should be computed
    method: str
        Method used
    min_n_rows: int
        Minimum number of rows for a KL estimation

    Returns
    -------
    pd.Series
        Kullback-Leibler divergence

    Raise
    -----
    AssertionError
        If the empirical distributions do not have enough samples to estimate a KL divergence.
        Consider using a larger dataset of lowering the parameter `min_n_rows`.
    """
    if method == "columnwise":
        cols_numerical = _get_numerical_features(df1)
        return columnwise_metric(
            df1[cols_numerical],
            df2[cols_numerical],
            df_mask[cols_numerical],
            kl_divergence_1D,
        )
    elif method == "gaussian":
        cols_numerical = _get_numerical_features(df1)
        return pattern_based_weighted_mean_metric(
            df1[cols_numerical],
            df2[cols_numerical],
            df_mask[cols_numerical],
            kl_divergence_gaussian,
            min_n_rows=min_n_rows,
        )
    elif method == "random_forest":
        return pattern_based_weighted_mean_metric(
            df1, df2, df_mask, kl_divergence_forest, min_n_rows=min_n_rows
        )
    else:
        raise AssertionError(
            f"The parameter of the function wasserstein_distance should be one of"
            f"the following: [`columnwise`, `gaussian`, `random_forest`], not `{method}`!"
        )


def distance_anticorr(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame) -> float:
    """Score based on the distance anticorrelation between two empirical distributions.
    The theoretical basis can be found on dcor documentation:
    https://dcor.readthedocs.io/en/latest/theory.html

    Parameters
    ----------
    df1 : pd.DataFrame
        Dataframe representing the first empirical distribution
    df2 : pd.DataFrame
        Dataframe representing the second empirical distribution
    df_mask: pd.DataFrame
        Mask indicating on what values the divergence should be computed

    Returns
    -------
    float
        Distance correlation score
    """
    df1 = df1[df_mask.any(axis=1)]
    df2 = df2[df_mask.any(axis=1)]
    return (1 - dcor.distance_correlation(df1.values, df2.values)) / 2


def pattern_based_weighted_mean_metric(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    df_mask: pd.DataFrame,
    metric: Callable,
    min_n_rows: int = 10,
    **kwargs,
) -> pd.Series:
    """Compute a mean score based on missing patterns.
    Note that for each pattern, a score is returned by the function metric.
    This code is based on https://www.statsmodels.org/

    Parameters
    ----------
    df1 : pd.DataFrame
        Dataframe representing the first empirical distribution, with nans
    df2 : pd.DataFrame
        Dataframe representing the second empirical distribution
    df_mask : pd.DataFrame
        Elements of the dataframes to compute on
    metric : Callable
        metric function
    min_n_rows : int, optional
        minimum number of row allowed for a pattern without nan, by default 10

    Returns
    -------
    pd.Series
        _description_
    """
    scores = []
    weights = []
    df1 = df1[df_mask.any(axis=1)]
    df2 = df2[df_mask.any(axis=1)]
    df_nan = df1.notna()
    max_num_row = 0
    for tup_pattern, df_nan_pattern in df_nan.groupby(df_nan.columns.tolist()):
        ind_pattern = df_nan_pattern.index
        df1_pattern = df1.loc[ind_pattern, list(tup_pattern)]
        max_num_row = max(max_num_row, len(df1_pattern))
        if not any(tup_pattern) or len(df1_pattern) < min_n_rows:
            continue
        df2_pattern = df2.loc[ind_pattern, list(tup_pattern)]
        df_mask_pattern = df_mask.loc[ind_pattern, list(tup_pattern)]
        weights.append(len(df1_pattern) / len(df1))
        scores.append(metric(df1_pattern, df2_pattern, df_mask_pattern, **kwargs))
    if len(scores) == 0:
        raise NotEnoughSamples(max_num_row, min_n_rows)
    return pd.Series(sum([s * w for s, w in zip(scores, weights)]), index=["All"])


def get_metric(name: str) -> Callable:
    dict_metrics: Dict[str, Callable] = {
        "mse": mean_squared_error,
        "rmse": root_mean_squared_error,
        "mae": mean_absolute_error,
        "wmape": weighted_mean_absolute_percentage_error,
        "wasserstein_columnwise": dist_wasserstein,
        "KL_columnwise": partial(kl_divergence, method="columnwise"),
        "KL_gaussian": partial(kl_divergence, method="gaussian"),
        "KL_forest": partial(kl_divergence, method="random_forest"),
        "ks_test": kolmogorov_smirnov_test,
        "correlation_diff": mean_difference_correlation_matrix_numerical_features,
        "energy": sum_energy_distances,
        "frechet": frechet_distance,
        "dist_corr_pattern": partial(
            pattern_based_weighted_mean_metric,
            metric=distance_anticorr,
        ),
    }
    return dict_metrics[name]



from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd

from qolmat.benchmark import hyperparameters, metrics
from qolmat.benchmark.missing_patterns import _HoleGenerator


class Comparator:
    """
    This class implements a comparator for evaluating different imputation methods.

    Parameters
    ----------
    dict_models: Dict[str, any]
        dictionary of imputation methods
    selected_columns: List[str]Œ
        list of column's names selected (all with at least one null value will be imputed)
    columnwise_evaluation : Optional[bool], optional
        whether the metric should be calculated column-wise or not, by default False
    dict_config_opti: Optional[Dict[str, Dict[str, Union[str, float, int]]]] = {}
        dictionary of search space for each implementation method. By default, the value is set to
        {}.
    max_evals: int = 10
        number of calls of the optimization algorithm
        10.
    """

    def __init__(
        self,
        dict_models: Dict[str, Any],
        selected_columns: List[str],
        generator_holes: _HoleGenerator,
        metrics: List = ["mae", "wmape", "KL_columnwise"],
        dict_config_opti: Optional[Dict[str, Any]] = {},
        metric_optim: str = "mse",
        max_evals: int = 10,
        verbose: bool = False,
    ):
        self.dict_imputers = dict_models
        self.selected_columns = selected_columns
        self.generator_holes = generator_holes
        self.metrics = metrics
        self.dict_config_opti = dict_config_opti
        self.metric_optim = metric_optim
        self.max_evals = max_evals
        self.verbose = verbose

    def get_errors(
        self,
        df_origin: pd.DataFrame,
        df_imputed: pd.DataFrame,
        df_mask: pd.DataFrame,
    ) -> pd.Series:
        """Functions evaluating the reconstruction's quality

        Parameters
        ----------
        signal_ref : pd.DataFrame
            reference/orginal signal
        signal_imputed : pd.DataFrame
            imputed signal

        Returns
        -------
        dictionary
            dictionay of results obtained via different metrics
        """
        dict_errors = {}
        for name_metric in self.metrics:
            fun_metric = metrics.get_metric(name_metric)
            dict_errors[name_metric] = fun_metric(df_origin, df_imputed, df_mask)
        errors = pd.concat(dict_errors.values(), keys=dict_errors.keys())
        return errors

    def evaluate_errors_sample(
        self,
        imputer: Any,
        df: pd.DataFrame,
        dict_config_opti_imputer: Dict[str, Any] = {},
        metric_optim: str = "mse",
    ) -> pd.Series:
        """Evaluate the errors in the cross-validation

        Parameters
        ----------
        tested_model : any
            imputation model
        df : pd.DataFrame
            dataframe to impute
        dict_config_opti_imputer : Dict
            search space for tested_model's hyperparameters
        metric_optim : str
            Loss function used when imputers undergo hyperparameter optimization

        Returns
        -------
        pd.DataFrame
            DataFrame with the errors for each metric (in column) and at each fold (in index)
        """
        list_errors = []
        df_origin = df[self.selected_columns].copy()
        for df_mask in self.generator_holes.split(df_origin):
            df_corrupted = df_origin.copy()
            df_corrupted[df_mask] = np.nan
            imputer_opti = hyperparameters.optimize(
                imputer,
                df,
                self.generator_holes,
                metric_optim,
                dict_config_opti_imputer,
                max_evals=self.max_evals,
                verbose=self.verbose,
            )
            df_imputed = imputer_opti.fit_transform(df_corrupted)
            subset = self.generator_holes.subset
            errors = self.get_errors(df_origin[subset], df_imputed[subset], df_mask[subset])
            list_errors.append(errors)
        df_errors = pd.DataFrame(list_errors)
        errors_mean = df_errors.mean(axis=0)

        return errors_mean

    def compare(
        self,
        df: pd.DataFrame,
    ):
        """Function to compare different imputation methods on dataframe df

        Parameters
        ----------
        df : pd.DataFrame
        verbose : bool, optional
            _description_, by default True
        Returns
        -------
        pd.DataFrame
            dataframe with imputation
        """

        dict_errors = {}

        for name, imputer in self.dict_imputers.items():
            dict_config_opti_imputer = self.dict_config_opti.get(name, {})

            try:
                dict_errors[name] = self.evaluate_errors_sample(
                    imputer, df, dict_config_opti_imputer, self.metric_optim
                )
                print(f"Tested model: {type(imputer).__name__}")
            except Exception as excp:
                print("Error while testing ", type(imputer).__name__)
                raise excp

        df_errors = pd.DataFrame(dict_errors)

        return df_errors



"""
Useful drawing functions
"""

from __future__ import annotations
from typing import Dict, List, Any, Optional, Tuple, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.ticker as plticker

import numpy as np
import pandas as pd
import scipy
from mpl_toolkits.axes_grid1 import make_axes_locatable

plt.rcParams["axes.spines.right"] = False
plt.rcParams["axes.spines.top"] = False
plt.rcParams["axes.grid"] = True
plt.rcParams["grid.alpha"] = 1
plt.rcParams["grid.color"] = "#cccccc"

tab10 = plt.get_cmap("tab10")


def plot_matrices(list_matrices: List[np.ndarray], title: Optional[str] = None) -> None:
    """Plot RPCA matrices

    Parameters
    ----------
    list_matrices : List[np.ndarray]
        List containing, in the right order, the observations matrix, the low-rank matrix and the
        sparse matrix
    title : Optional[str], optional
        if present, title of the saved figure, by default None
    """

    suptitles = ["Observations", "Low-rank", "Sparse"]

    fig, ax = plt.subplots(1, 3, figsize=(10, 3))

    for i, (m, t) in enumerate(zip(list_matrices, suptitles)):
        if i != 2:
            im = ax[i].imshow(m, aspect="auto")
        else:
            m = ax[i].imshow(m, aspect="auto")
        divider = make_axes_locatable(ax[i])
        cax = divider.append_axes("right", size="5%", pad=0.05)
        fig.colorbar(im, cax=cax, orientation="vertical")
        ax[i].set_title(t, fontsize=16)

    plt.tight_layout()
    if title:
        plt.savefig(f"../figures/{title}.png", transparent=True)

    plt.show()


def plot_signal(
    list_signals: List[List],
    title: Optional[str] = None,
    ylabel: Optional[str] = None,
    dates: Optional[List] = None,
) -> None:
    """Plot RPCA results for time series

    Parameters
    ----------
    list_signals : List[List]
        List containing, in the right order, the  observed time series, the cleaned signal and
        the anomalies
    title : Optional[str], optional
        if present, title of the saved figure, by default None
    ylabel : Optional[str], optional
        ylabel, by default None
    dates : Optional[List], optional
        dates of the time series (xlabel), by default None
    """

    suptitles = ["Observations", "Cleaned", "Anomalies"]
    colors = ["black", "darkblue", "crimson"]
    fontsize = 15

    if dates is None:
        dates = list(range(len(list_signals[0])))

    fig, ax = plt.subplots(3, 1, sharex=True, figsize=(15, 6))
    for i, (r, c, t) in enumerate(zip(list_signals, colors, suptitles)):
        ax[i].plot(dates, r, color=c)
        ax[i].set_title(t, fontsize=fontsize)
        if ylabel:
            ax[i].set_ylabel(ylabel, fontsize=fontsize)

    fig.align_ylabels()
    plt.tight_layout()
    if title:
        plt.savefig(f"../figures/{title}.png", transparent=True)
    plt.show()


def plot_images(
    M: np.ndarray,
    A: np.ndarray,
    E: np.ndarray,
    index_array: List[int],
    dims: Tuple[int, int],
    filename: Optional[str] = None,
) -> None:
    """Plot multiple images in 3 columns for original, background and "foreground"

    Parameters
    ----------
    M : np.ndarray
        orginal array
    A : np.ndarray
        background array
    E : np.ndarray
        foreground/moving object array
    index_array : List[int]
        indices of the plotted frames
    dims : Tuple[int, int]
        dimensions of the reduction
    filename : Optional[str], optional
        filename for saving figure, by default None
    """

    f = plt.figure(figsize=(15, 10))
    r = len(index_array)

    for k, i in enumerate(index_array):
        for j, mat in enumerate([M, A, E]):
            sp = f.add_subplot(r, 3, 3 * k + j + 1)
            sp.set_xticks([])
            sp.set_yticks([])
            pixels = mat[:, i]
            if isinstance(pixels, scipy.sparse.csr_matrix):
                pixels = pixels.todense()
            sp.imshow(np.reshape(pixels, dims), cmap="gray")

            if j == 0:
                sp.set_ylabel(f"Frame {i}", fontsize=25)

            if k == 0:
                if j == 0:
                    sp.set_title("Original", fontsize=25)
                elif j == 1:
                    sp.set_title("Background", fontsize=25)
                else:
                    sp.set_title("Moving objects", fontsize=25)

    if filename:
        plt.savefig(f"../figures/{filename}.png", transparent=True)

    plt.tight_layout()
    plt.show()


def make_ellipses(
    X: np.ndarray,
    ax: mpl.axes.Axes,
    color: Union[str, Tuple[float, float, float]],
):
    """Draw ellipses on a figure

    Parameters
    ----------
    X : np.ndarray
        array for the ellipse
    ax : matplotlib.axes._subplots.AxesSubplot
        matplotlib ax handles
    color : Union[str, Tuple[float, float, float]]
        ellipse's color
    """
    covariances = np.cov(X)  # gmm.covariances_[0] # [n][:2, :2]
    v, w = np.linalg.eigh(covariances)
    u = w[0] / np.linalg.norm(w[0])
    angle = np.arctan2(u[1], u[0])
    angle = 180 * angle / np.pi  # convert to degrees
    center = X.mean(axis=0)  # .means_[0]
    v[v < 0] = np.nan
    v = 2.0 * np.sqrt(2.0) * np.sqrt(v)
    ell = mpl.patches.Ellipse(center, v[0], v[1], angle=180 + angle, color=color)
    ell.set_clip_box(ax.bbox)
    ell.set_alpha(0.5)
    ax.add_artist(ell)
    ax.set_aspect("equal", "datalim")


def compare_covariances(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    col_x: str,
    col_y: str,
    ax: mpl.axes._subplots.AxesSubplot,
    label: str = "",
    color: Optional[Union[str, Tuple[float, float, float]]] = None,
):
    """
    Covariance plot: scatter plot with ellipses

    Parameters
    ----------
    df1 : pd.DataFrame
        dataframe with raw data
    df2 : pd.DataFrame
        dataframe with imputations
    col_x : str
        variable x, column's name of dataframe df1 to compare with
    col_y : str
        variable y, column's name of dataframe df2 to compare with
    ax : matplotlib.axes._subplots.AxesSubplot
        matplotlib ax handles
    """
    if color is None:
        color = tab10(0)
    ax.scatter(df2[col_x], df2[col_y], marker=".", color=color, s=0.2, label=label)
    ax.scatter(df1[col_x], df1[col_y], marker=".", color="black", s=0.2)
    make_ellipses(df1[[col_x, col_y]], ax, "black")
    make_ellipses(df2[[col_x, col_y]], ax, color)
    ax.set_xlabel(col_x)
    ax.set_ylabel(col_y)


def multibar(
    df: pd.DataFrame,
    ax: Optional[mpl.axes.Axes] = None,
    orientation: str = "vertical",
    colors: Any = None,
    decimals: float = 0,
):
    """Create a multi-bar graph to represent the values of the different dataframe columns.

    Parameters
    ----------
    df : pd.DataFrame
        contain the dataframe
    ax : Any, optional
        matplotlib ax handles, by default None
    orientation : str, optional
        orientation of plot, by default "vertical"
    colors : str, optional
        color in multibar plot, by default None
    decimals : float, optional
        the decimals numbers, by default 0
    """

    if ax is None:
        ax = plt.gca()
        if colors is None:
            colors = tab10
    x = np.arange(len(df))  # the label locations
    n_columns = len(df.columns)
    width_tot = 0.8
    width_col = width_tot / n_columns  # the width of the bars

    for i_column, column in enumerate(df.columns):
        color_col = colors(i_column % 10)
        dx = width_tot * (-0.5 + float(i_column) / n_columns)
        if orientation == "horizontal":
            rect = ax.barh(
                x + dx,
                df[column],
                width_col,
                label=column,
                align="edge",
                color=color_col,
            )
            plt.yticks(x, df.index)
        else:
            rect = ax.bar(
                x + dx,
                df[column],
                width_col,
                label=column,
                align="edge",
                color=color_col,
            )
            plt.xticks(x, df.index)
        ax.bar_label(rect, padding=3, fmt=f"%.{decimals}f")

    plt.legend(loc=(1, 0))


def plot_imputations(df: pd.DataFrame, dict_df_imputed: Dict[str, pd.DataFrame]):
    """Plot original and imputed dataframes for each imputers

    Parameters
    ----------
    df : pd.DataFrame
        original dataframe
    dict_df_imputed : Dict[str, pd.DataFrame]
        dictionnary of imputed dataframe for each imputers
    """
    n_columns = len(df.columns)
    n_imputers = len(dict_df_imputed)

    fig = plt.figure(figsize=(8 * n_columns, 6 * n_imputers))
    i_plot = 1
    for name_imputer, df_imputed in dict_df_imputed.items():
        for col in df:
            ax = fig.add_subplot(n_imputers, n_columns, i_plot)
            values_orig = df[col]

            plt.plot(values_orig, ".", color="black", label="original")
            values_imp = df_imputed[col].copy()
            values_imp[values_orig.notna()] = np.nan
            plt.plot(values_imp, ".", color=tab10(0), label=name_imputer, alpha=1)
            plt.ylabel(col, fontsize=16)
            if i_plot % n_columns == 0:
                plt.legend(loc=[1, 0], fontsize=18)
            loc = plticker.MultipleLocator(base=2 * 365)
            ax.xaxis.set_major_locator(loc)
            ax.tick_params(axis="both", which="major", labelsize=17)
            i_plot += 1
    plt.show()



from typing import Tuple
import numpy as np
import pandas as pd
import pytest
from numpy.typing import NDArray
from pytest_mock.plugin import MockerFixture
from qolmat.imputations.rpca.rpca import RPCA

# X_incomplete = np.array([[1, np.nan], [4, 2], [np.nan, 4]])


# X_exp_nrows_1_prepare_data = np.array([1.0, np.nan, 4.0, 2.0, np.nan, 4.0])
# X_exp_nrows_6_prepare_data = np.concatenate(
#     [X_incomplete.reshape(-1, 6).flatten(), np.ones((1, 94)).flatten() * np.nan]
# )

# period = 100
# max_iter = 256
# mu = 0.5
# tau = 0.5
# lam = 1


class RPCAMock(RPCA):
    def __init__(self):
        super().__init__()

    def decompose_rpca(self, D: NDArray, Omega: NDArray) -> Tuple[NDArray, NDArray]:
        self.call_count = 1
        return D, D


X_incomplete = np.array([[1, np.nan], [4, 2], [np.nan, 4]])


def test_rpca_decompose_rpca_signal() -> None:
    rpca = RPCAMock()
    M, A = rpca.decompose_rpca_signal(X_incomplete)
    assert M.shape == X_incomplete.shape
    assert A.shape == X_incomplete.shape
    assert rpca.call_count == 1



from __future__ import annotations

from typing import Tuple, Union

import numpy as np
from numpy.typing import NDArray
from sklearn.base import BaseEstimator, TransformerMixin

from qolmat.utils import utils


class RPCA(BaseEstimator, TransformerMixin):
    """
    This class is the root class of the RPCA methods.

    Parameters
    ----------
    period: Optional[int]
        Number of rows of the array if the array is
        1D and reshaped into a 2D array, by default `None`.
    max_iter: int
        maximum number of iterations of the
        alternating direction method of multipliers,
        by default 1e4.
    tol: float
        Tolerance for stopping criteria, by default 1e-6
    verbose: bool
        default `False`
    """

    def __init__(
        self,
        period: int = 1,
        max_iterations: int = int(1e4),
        tol: float = 1e-6,
        random_state: Union[None, int, np.random.RandomState] = None,
        verbose: bool = True,
    ) -> None:
        self.period = period
        self.max_iterations = max_iterations
        self.tol = tol
        self.random_state = random_state
        self.verbose = verbose

    def decompose_rpca_signal(
        self,
        X: NDArray,
    ) -> Tuple[NDArray, NDArray]:
        """
        Compute the noisy RPCA with L1 or L2 time penalisation

        Parameters
        ----------
        X : NDArray
            Observations

        Returns
        -------
        M: NDArray
            Low-rank signal
        A: NDArray
            Anomalies
        """

        D = utils.prepare_data(X, self.period)
        Omega = ~np.isnan(D)
        # D_proj = rpca_utils.impute_nans(D_init, method="median")
        D = utils.linear_interpolation(D)
        n_rows, n_cols = D.shape
        if n_rows == 1 or n_cols == 1:
            return D, np.full_like(D, 0)

        M, A = self.decompose_rpca(D, Omega)

        M_final = utils.get_shape_original(M, X.shape)
        A_final = utils.get_shape_original(A, X.shape)

        return M_final, A_final



import numpy as np
import pandas as pd
import pytest

from qolmat.imputations import imputers_pytorch
from qolmat.utils.exceptions import PyTorchExtraNotInstalled

# from __future__ import annotations

try:
    import torch as nn
except ModuleNotFoundError:
    raise PyTorchExtraNotInstalled

df_incomplete = pd.DataFrame(
    {
        "col1": [np.nan, 15, np.nan, 23, 33],
        "col2": [69, 76, 74, 80, 78],
        "col3": [174, 166, 182, 177, np.nan],
        "col4": [9, 12, 11, 12, 8],
        "col5": [93, 75, np.nan, 12, np.nan],
    }
)


@pytest.mark.parametrize("df", [df_incomplete])
def test_ImputerRegressorTorch_fit_transform(df: pd.DataFrame) -> None:
    nn.manual_seed(42)
    if nn.cuda.is_available():
        nn.cuda.manual_seed(42)
    estimator = imputers_pytorch.build_mlp_example(input_dim=2, list_num_neurons=[64, 32])
    imputer = imputers_pytorch.ImputerRegressorPyTorch(
        estimator=estimator, handler_nan="column", epochs=10
    )

    result = imputer.fit_transform(df)
    expected = pd.DataFrame(
        {
            "col1": [26.596, 15.0, 28.483, 23.0, 33.0],
            "col2": [69.0, 76.0, 74.0, 80.0, 78.0],
            "col3": [174.0, 166.0, 182.0, 177.0, 178.786],
            "col4": [9.0, 12.0, 11.0, 12.0, 8.0],
            "col5": [93.0, 75.0, 48.626, 12.0, 51.389],
        }
    )
    np.testing.assert_allclose(result, expected, atol=1e-3)



import datetime

import numpy as np
import pandas as pd
import pytest
from pytest_mock.plugin import MockerFixture

from qolmat.utils import data

columns = ["No", "year", "month", "day", "hour", "a", "b", "wd"]
df_beijing = pd.DataFrame(
    [
        [1, 2013, 3, 1, 0, 1, 2, "NW"],
        [2, 2014, 3, 1, 0, 3, np.nan, "NW"],
        [3, 2015, 3, 1, 0, np.nan, 6, "NW"],
    ],
    columns=columns,
)
index_preprocess_beijing = pd.MultiIndex.from_tuples(
    [
        ("Beijing", datetime.datetime(2013, 3, 1)),
        ("Beijing", datetime.datetime(2014, 3, 1)),
        ("Beijing", datetime.datetime(2015, 3, 1)),
    ],
    names=["station", "datetime"],
)
df_preprocess_beijing = pd.DataFrame(
    [[1, 2], [3, np.nan], [np.nan, 6]], columns=["a", "b"], index=index_preprocess_beijing
)

columns = ["No", "year", "month", "day", "hour", "a", "b", "wd", "station"]
df_offline = pd.DataFrame(
    [
        [1, 2013, 3, 1, 0, 1, 2, "NW", "Gucheng"],
        [2, 2014, 3, 1, 0, 3, np.nan, "NW", "Gucheng"],
        [3, 2015, 3, 1, 0, np.nan, 6, "NW", "Gucheng"],
    ],
    columns=columns,
)

df_monach_weather = pd.DataFrame(
    {
        "series_name": ["T1", "T2", "T3", "T4", "T5"],
        "series_type": ["rain", "preasure", "temperature", "humidity", "sun"],
        "series_value": [
            [1.0, 2.0, 3.0],
            [4.0, 5.0, 6.0],
            [5.0, 4.0, 3.0],
            [2.0, 1.0, 4.0],
            [1.0, 4.0, 6.0],
        ],
    }
)

df_monach_weather_preprocess = pd.DataFrame(
    [
        [1.0, 4.0, 5.0, 2.0, 1.0],
        [2.0, 5.0, 4.0, 1.0, 4.0],
        [3.0, 6.0, 3.0, 4.0, 6.0],
    ],
    columns=["T1 rain", "T2 preasure", "T3 temperature", "T4 humidity", "T5 sun"],
    index=pd.date_range(start="2010-01-01", periods=3, freq="1D"),
)


df_monach_elec = pd.DataFrame(
    {
        "series_name": ["T1", "T2", "T3", "T4", "T5"],
        "state": ["NSW", "VIC", "QUN", "SA", "TAS"],
        "start_timestamp": [
            pd.Timestamp("2002-01-01"),
            pd.Timestamp("2002-01-01"),
            pd.Timestamp("2002-01-01"),
            pd.Timestamp("2002-01-01"),
            pd.Timestamp("2002-01-01"),
        ],
        "series_value": [
            [5714.0, 5360.0, 5014.0],
            [3535.0, 3383.0, 3655.0],
            [3382.0, 3288.0, 3172.0],
            [1191.0, 1219.0, 1119.0],
            [315.0, 306.0, 305.0],
        ],
    }
)

df_monach_elec_preprocess = pd.DataFrame(
    [
        [5714.0, 3535.0, 3382.0, 1191.0, 315.0],
        [5360.0, 3383.0, 3288.0, 1219.0, 306.0],
        [5014.0, 3655.0, 3172.0, 1119.0, 305.0],
    ],
    columns=["T1 NSW", "T2 VIC", "T3 QUN", "T4 SA", "T5 TAS"],
    index=pd.date_range(start="2002-01-01", periods=3, freq="30T"),
)

index_preprocess_offline = pd.MultiIndex.from_tuples(
    [
        ("Gucheng", datetime.datetime(2013, 3, 1)),
        ("Gucheng", datetime.datetime(2014, 3, 1)),
        ("Gucheng", datetime.datetime(2015, 3, 1)),
    ],
    names=["station", "datetime"],
)
df_preprocess_offline = pd.DataFrame(
    [[1, 2], [3, np.nan], [np.nan, 6]], columns=["a", "b"], index=index_preprocess_offline
)


urllink = "https://archive.ics.uci.edu/ml/machine-learning-databases/00501/"
zipname = "PRSA2017_Data_20130301-20170228"


# @pytest.mark.parametrize("zipname, urllink", [(zipname, urllink)])
# def test_utils_data_download_data(zipname: str, urllink: str, mocker: MockerFixture) -> None:
#     mocker.patch("urllib.request.urlretrieve")
#     mocker.patch("zipfile.ZipFile")
#     list_df_result = data.download_data(zipname, urllink)


@pytest.mark.parametrize(
    "name_data, df",
    [
        ("Beijing", df_beijing),
        ("Beijing_offline", df_offline),
        ("Monach_weather", df_monach_weather),
        ("Monach_electricity_australia", df_monach_elec),
        ("Artificial", None),
        ("Bug", None),
    ],
)
def test_utils_data_get_data(name_data: str, df: pd.DataFrame, mocker: MockerFixture) -> None:
    mock_download = mocker.patch("qolmat.utils.data.download_data", return_value=[df])
    mocker.patch(
        "qolmat.utils.data.preprocess_data_beijing_offline", return_value=df_preprocess_offline
    )
    mocker.patch("qolmat.utils.data.preprocess_data_beijing", return_value=df_preprocess_beijing)
    mock_get = mocker.patch("qolmat.utils.data.get_dataframes_in_folder", return_value=[df])
    try:
        df_result = data.get_data(name_data=name_data)
    except ValueError:
        assert name_data not in [
            "Beijing",
            "Beijing_offline",
            "Monach_weather",
            "Monach_electricity_australia",
            "Artificial",
        ]
        np.testing.assert_raises(ValueError, data.get_data, name_data)
        return

    if name_data == "Beijing":
        assert mock_download.call_count == 1
        pd.testing.assert_frame_equal(df_result, df_preprocess_beijing)
    elif name_data == "Beijing_offline":
        assert mock_download.call_count == 0
        assert mock_get.call_count == 1
        pd.testing.assert_frame_equal(df_result, df_preprocess_offline)
    elif name_data == "Artificial":
        expected_columns = ["signal", "X", "A", "E"]
        assert isinstance(df_result, pd.DataFrame)
        assert df_result.columns.tolist() == expected_columns
    elif name_data == "Monach_weather":
        assert mock_download.call_count == 1
        print(df_result)
        pd.testing.assert_frame_equal(df_result, df_monach_weather_preprocess)
    elif name_data == "Monach_electricity_australia":
        assert mock_download.call_count == 1
        print(df_result)
        pd.testing.assert_frame_equal(df_result, df_monach_elec_preprocess)
    else:
        assert False


@pytest.mark.parametrize("df", [df_offline])
def test_utils_data_preprocess_data_beijing_offline(df: pd.DataFrame) -> None:
    result = data.preprocess_data_beijing_offline(df)
    pd.testing.assert_frame_equal(result, df_preprocess_offline, atol=1e-3)


@pytest.mark.parametrize("df", [df_preprocess_offline])
def test_utils_data_add_holes(df: pd.DataFrame) -> None:
    df_out = data.add_holes(df, 0.0, 1)
    assert df_out.isna().sum().sum() == 2
    df_out = data.add_holes(df, 1.0, 1)
    assert df_out.isna().sum().sum() > 2


@pytest.mark.parametrize("name_data", ["Beijing"])
def test_utils_data_get_data_corrupted(name_data: str, mocker: MockerFixture) -> None:
    mock_download = mocker.patch("qolmat.utils.data.download_data", return_value=[df_beijing])
    mocker.patch("qolmat.utils.data.preprocess_data_beijing", return_value=df_preprocess_beijing)
    df_out = data.get_data_corrupted()
    df_result = pd.DataFrame(
        [[1, 2], [np.nan, np.nan], [np.nan, 6]], columns=["a", "b"], index=index_preprocess_beijing
    )
    assert mock_download.call_count == 1
    pd.testing.assert_frame_equal(df_result, df_out)


@pytest.mark.parametrize("df", [df_preprocess_beijing])
def test_utils_data_add_station_features(df: pd.DataFrame) -> None:
    columns_out = ["a", "b"] + ["station=Beijing"]
    expected = pd.DataFrame(
        [
            [1, 2, 1.0],
            [3, np.nan, 1.0],
            [np.nan, 6, 1.0],
        ],
        columns=columns_out,
        index=index_preprocess_beijing,
    )
    result = data.add_station_features(df)
    pd.testing.assert_frame_equal(result, expected, atol=1e-3)


@pytest.mark.parametrize("df", [df_preprocess_beijing])
def test_utils_data_add_datetime_features(df: pd.DataFrame) -> None:
    columns_out = ["a", "b"] + ["time_cos"]
    expected = pd.DataFrame(
        [
            [1, 2, 0.512],
            [3, np.nan, 0.512],
            [np.nan, 6, 0.512],
        ],
        columns=columns_out,
        index=index_preprocess_beijing,
    )
    result = data.add_datetime_features(df)
    pd.testing.assert_frame_equal(result, expected, atol=1e-3)



from typing import Any, Dict

import numpy as np
import pandas as pd
import pytest
from sklearn.utils.estimator_checks import check_estimator, parametrize_with_checks
from qolmat.benchmark.hyperparameters import HyperValue
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import ExtraTreesRegressor


from qolmat.imputations import imputers

df_complete = pd.DataFrame({"col1": [0, 1, 2, 3, 4], "col2": [-1, 0, 0.5, 1, 1.5]})

df_incomplete = pd.DataFrame(
    {"col1": [0, np.nan, 2, 3, np.nan], "col2": [-1, np.nan, 0.5, np.nan, 1.5]}
)

df_timeseries = pd.DataFrame(
    pd.DataFrame(
        {
            "col1": [i for i in range(20)],
            "col2": [0, np.nan, 2, np.nan, 2] + [i for i in range(5, 20)],
        },
        index=pd.date_range("2023-04-17", periods=20, freq="D"),
    )
)

df_groups = pd.DataFrame(
    {
        "col1": [1, 1, 0, 1],
        "col2": [1, np.nan, 0, 3],
    }
)


def test_hyperparameters_get_hyperparameters() -> None:
    imputer = imputers.ImputerKNN(n_neighbors=3)
    hyperparams = imputer.get_hyperparams("col")

    assert hyperparams == {"n_neighbors": 3, "weights": "distance"}


hyperparams_global = {
    "lam/col1": 4.7,
    "lam/col2": 1.5,
    "tol": 0.07,
    "max_iterations": 100,
    "norm": "L1",
}

expected1 = {
    "lam": 4.7,
    "tau": None,
    "mu": None,
    "rank": None,
    "list_etas": (),
    "list_periods": (),
    "tol": 0.07,
    "norm": "L1",
    "max_iterations": 100,
    "period": 1,
}

expected2 = {
    "lam": 1.5,
    "tau": None,
    "mu": None,
    "rank": None,
    "list_etas": (),
    "list_periods": (),
    "tol": 0.07,
    "norm": "L1",
    "max_iterations": 100,
    "period": 1,
}


@pytest.mark.parametrize("col, expected", [("col1", expected1), ("col2", expected2)])
def test_hyperparameters_get_hyperparameters_modified(
    col: str, expected: Dict[str, HyperValue]
) -> None:
    imputer = imputers.ImputerRPCA()
    for key, val in hyperparams_global.items():
        setattr(imputer, key, val)
    imputer.imputer_params = tuple(set(imputer.imputer_params) | set(hyperparams_global.keys()))
    hyperparams = imputer.get_hyperparams(col)

    assert hyperparams == expected


@pytest.mark.parametrize(
    "imputer",
    [
        imputers.ImputerMean(),
        imputers.ImputerMedian(),
        imputers.ImputerMode(),
        imputers.ImputerShuffle(),
        imputers.ImputerLOCF(),
        imputers.ImputerNOCB(),
    ],
)
@pytest.mark.parametrize(
    "df", [pd.DataFrame({"col1": [np.nan, np.nan, np.nan], "col2": [1, 2, 3]})]
)
def test_Imputer_fit_transform_on_nan_column(df: pd.DataFrame, imputer: imputers._Imputer) -> None:
    np.testing.assert_raises(ValueError, imputer.fit_transform, df)


@pytest.mark.parametrize("df", ["string", [1, 2, 3]])
def test_fit_transform_not_on_pandas(df: Any) -> None:
    imputer = imputers.ImputerMean()
    np.testing.assert_raises(ValueError, imputer.fit_transform, df)


@pytest.mark.parametrize("df", [df_groups])
def test_fit_transform_on_grouped(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerMean(groups=("col1",))
    result = imputer.fit_transform(df)
    expected = pd.DataFrame(
        {
            "col1": [1, 1, 0, 1],
            "col2": [1, 2, 0, 3],
        }
    )
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [df_incomplete])
@pytest.mark.parametrize("df_oracle", [df_complete])
def test_ImputerOracle_fit_transform(df: pd.DataFrame, df_oracle: pd.DataFrame) -> None:
    imputer = imputers.ImputerOracle()
    imputer.set_solution(df_oracle)
    result = imputer.fit_transform(df)
    expected = df_oracle
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [df_incomplete])
def test_ImputerMean_fit_transform(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerMean()
    result = imputer.fit_transform(df)
    expected = pd.DataFrame(
        {"col1": [0, 5 / 3, 2, 3, 5 / 3], "col2": [-1, 1 / 3, 0.5, 1 / 3, 1.5]}
    )
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [df_incomplete])
def test_ImputerMedian_fit_transform(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerMedian()
    result = imputer.fit_transform(df)
    expected = pd.DataFrame({"col1": [0, 2, 2, 3, 2], "col2": [-1, 0.5, 0.5, 0.5, 1.5]})
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [df_incomplete])
def test_ImputerMode_fit_transform(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerMode()
    result = imputer.fit_transform(df)
    expected = pd.DataFrame({"col1": [0, 0, 2, 3, 0], "col2": [-1, -1, 0.5, -1, 1.5]})
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [pd.DataFrame({"col1": [1, 1, np.nan]})])
def test_ImputerShuffle_fit_transform1(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerShuffle()
    result = imputer.fit_transform(df)
    expected = pd.DataFrame({"col1": [1, 1, 1]})
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [df_incomplete])
def test_ImputerShuffle_fit_transform2(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerShuffle(random_state=42)
    result = imputer.fit_transform(df)
    expected = pd.DataFrame({"col1": [0, 3, 2, 3, 0], "col2": [-1, 1.5, 0.5, 1.5, 1.5]})
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [df_incomplete])
def test_ImputerLOCF_fit_transform(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerLOCF()
    result = imputer.fit_transform(df)
    expected = pd.DataFrame({"col1": [0, 0, 2, 3, 3], "col2": [-1, -1, 0.5, 0.5, 1.5]})
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [df_incomplete])
def test_ImputerNOCB_fit_transform(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerNOCB()
    result = imputer.fit_transform(df)
    expected = pd.DataFrame({"col1": [0, 2, 2, 3, 3], "col2": [-1, 0.5, 0.5, 1.5, 1.5]})
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [df_incomplete])
def test_ImputerInterpolation_fit_transform(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerInterpolation()
    result = imputer.fit_transform(df)
    expected = pd.DataFrame({"col1": [0, 1, 2, 3, 3], "col2": [-1, -0.25, 0.5, 1, 1.5]})
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [df_timeseries])
def test_ImputerResiduals_fit_transform(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerResiduals(period=7)
    result = imputer.fit_transform(df)
    expected = pd.DataFrame(
        {
            "col1": [i for i in range(20)],
            "col2": [0, 0.953, 2, 2.061, 2] + [i for i in range(5, 20)],
        },
        index=pd.date_range("2023-04-17", periods=20, freq="D"),
    )
    np.testing.assert_allclose(result, expected, atol=1e-3)


@pytest.mark.parametrize("df", [df_incomplete])
def test_ImputerKNN_fit_transform(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerKNN(n_neighbors=2)
    result = imputer.fit_transform(df)
    expected = pd.DataFrame(
        {
            "col1": [0, 5 / 3, 2, 3, 1.4285714285714286],
            "col2": [-1, 1 / 3, 0.5, 1 / 8, 1.5],
        }
    )
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [df_incomplete])
def test_ImputerMICE_fit_transform(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerMICE(
        estimator=ExtraTreesRegressor(),
        random_state=42,
        sample_posterior=False,
        max_iter=100,
    )
    result = imputer.fit_transform(df)
    expected = pd.DataFrame(
        {
            "col1": [0, 3, 2, 3, 3],
            "col2": [-1, 1.5, 0.5, 1.5, 1.5],
        }
    )
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [df_incomplete])
def test_ImputerRegressor_fit_transform(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerRegressor(estimator=ExtraTreesRegressor())
    result = imputer.fit_transform(df)
    expected = pd.DataFrame(
        {
            "col1": [0, 2, 2, 3, 2],
            "col2": [-1, 0.5, 0.5, 0.5, 1.5],
        }
    )
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("df", [df_timeseries])
def test_ImputerRPCA_fit_transform(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerRPCA(columnwise=False, max_iterations=100, tau=1, lam=0.3)
    result = imputer.fit_transform(df)
    expected = pd.DataFrame(
        {
            "col1": [i for i in range(20)],
            "col2": [0, 1, 2, 2, 2] + [i for i in range(5, 20)],
        }
    )
    result = np.around(result)
    np.testing.assert_allclose(result, expected, atol=1e-2)


@pytest.mark.parametrize("df", [df_timeseries])
def test_ImputerEM_fit_transform(df: pd.DataFrame) -> None:
    imputer = imputers.ImputerEM(method="sample", random_state=42)
    result = imputer.fit_transform(df)
    expected = pd.DataFrame(
        {
            "col1": [i for i in range(20)],
            "col2": [0, 21.074, 2, -4.262, 2] + [i for i in range(5, 20)],
        }
    )
    np.testing.assert_allclose(result, expected, atol=1e-2)


index_grouped = pd.MultiIndex.from_product([["a", "b"], range(4)], names=["group", "date"])
dict_values = {"col1": [0, np.nan, 0, 0, 1, 1, 1, 1], "col2": np.arange(8)}
df_grouped = pd.DataFrame(dict_values, index=index_grouped)

list_imputers = [
    imputers.ImputerMean(groups=("group",)),
    imputers.ImputerMedian(groups=("group",)),
    imputers.ImputerMode(groups=("group",)),
    imputers.ImputerShuffle(groups=("group",)),
    imputers.ImputerLOCF(groups=("group",)),
    imputers.ImputerNOCB(groups=("group",)),
    imputers.ImputerInterpolation(groups=("group",)),
    imputers.ImputerResiduals(groups=("group",), period=2),
    imputers.ImputerKNN(groups=("group",)),
    imputers.ImputerMICE(groups=("group",)),
    imputers.ImputerRegressor(groups=("group",), estimator=LinearRegression()),
    imputers.ImputerRPCA(groups=("group",)),
    imputers.ImputerEM(groups=("group",)),
]


@pytest.mark.parametrize("imputer", list_imputers)
def test_models_fit_transform_grouped(imputer):
    # imputer = imputers.ImputerEM(groups=("group",), method="sample", random_state=42)
    result = imputer.fit_transform(df_grouped)
    expected = df_grouped.fillna(0)
    np.testing.assert_allclose(result, expected)


@parametrize_with_checks(
    [
        imputers._Imputer(),
        imputers.ImputerOracle(),
        imputers.ImputerMean(),
        imputers.ImputerMedian(),
        imputers.ImputerMode(),
        imputers.ImputerShuffle(),
        imputers.ImputerLOCF(),
        imputers.ImputerNOCB(),
        imputers.ImputerInterpolation(),
        imputers.ImputerResiduals(period=2),
        imputers.ImputerKNN(),
        imputers.ImputerMICE(),
        imputers.ImputerRegressor(),
        imputers.ImputerRPCA(tau=0, lam=0),
        imputers.ImputerEM(),
    ]
)
def test_sklearn_compatible_estimator(estimator: imputers._Imputer, check: Any) -> None:
    """Check compatibility with sklearn, using sklearn estimator checks API."""
    check(estimator)



import sys
import numpy as np
from numpy.typing import NDArray
import pandas as pd
import pytest
from qolmat.utils import utils
from pytest_mock.plugin import MockerFixture
from io import StringIO

from qolmat.utils.exceptions import NotDimension2, SignalTooShort


df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6], "C": [7, 8, 9]})


@pytest.mark.parametrize("iteration, total", [(1, 1)])
def test_utils_utils_display_progress_bar(iteration: int, total: int, capsys) -> None:
    captured_output = StringIO()
    sys.stdout = captured_output
    utils.progress_bar(
        iteration, total, prefix="Progress", suffix="Complete", decimals=1, length=2, fill="█"
    )
    captured_output.seek(0)
    output = captured_output.read().strip()
    sys.stdout = sys.__stdout__

    output_expected = "Progress |██| 100.0% Complete"
    assert output == output_expected


@pytest.mark.parametrize("values, lag_max", [(pd.Series([1, 2, 3, 4, 5]), 3)])
def test_utils_utils_acf(values, lag_max):
    result = utils.acf(values, lag_max)
    result_expected = pd.Series([1.0, 1.0, 1.0])
    pd.testing.assert_series_equal(result, result_expected, atol=0.001)


X_incomplete = np.array(
    [
        [1, np.nan, 3, 2, np.nan],
        [7, 2, np.nan, 1, 1],
        [np.nan, 4, 3, np.nan, 5],
        [np.nan, 4, 3, 5, 5],
        [4, 4, 3, np.nan, 5],
    ]
)
X_exp_mean = np.array(
    [
        [1.0, 3.5, 3.0, 2.0, 4.0],
        [7.0, 2.0, 3.0, 1.0, 1.0],
        [4.0, 4.0, 3.0, 2.66666667, 5.0],
        [4.0, 4.0, 3.0, 5.0, 5.0],
        [4.0, 4.0, 3.0, 2.66666667, 5.0],
    ]
)
X_exp_median = np.array(
    [
        [1.0, 4.0, 3.0, 2.0, 5.0],
        [7.0, 2.0, 3.0, 1.0, 1.0],
        [4.0, 4.0, 3.0, 2.0, 5.0],
        [4.0, 4.0, 3.0, 5.0, 5.0],
        [4.0, 4.0, 3.0, 2.0, 5.0],
    ]
)
X_exp_zeros = np.array(
    [
        [1.0, 0.0, 3.0, 2.0, 0.0],
        [7.0, 2.0, 0.0, 1.0, 1.0],
        [0.0, 4.0, 3.0, 0.0, 5.0],
        [0.0, 4.0, 3.0, 5.0, 5.0],
        [4.0, 4.0, 3.0, 0.0, 5.0],
    ]
)


@pytest.mark.parametrize("X", [X_incomplete])
@pytest.mark.parametrize(
    "method, X_expected", [("mean", X_exp_mean), ("median", X_exp_median), ("zeros", X_exp_zeros)]
)
def test_utils_utils_impute_nans(X: NDArray, method: str, X_expected: NDArray):
    result = utils.impute_nans(M=X, method=method)
    np.testing.assert_allclose(result, X_expected)


@pytest.mark.parametrize("X", [X_incomplete])
def test_utils_linear_interpolation(X: NDArray):
    result = utils.linear_interpolation(X_incomplete)
    expected = np.array(
        [
            [1, 2, 3, 2, 2],
            [7, 2, 1.5, 1, 1],
            [4, 4, 3, 4, 5],
            [4, 4, 3, 5, 5],
            [4, 4, 3, 4, 5],
        ]
    )
    np.testing.assert_allclose(result, expected)


signal = np.array([1, 4, np.nan, 3, 2])
X_expected3 = np.array(
    [
        [1.0, np.nan, 3.0, 2.0, np.nan, 7.0, 2.0, np.nan, 1.0],
        [1.0, np.nan, 4.0, 3.0, np.nan, 5.0, np.nan, 4.0, 3.0],
        [5.0, 5.0, 4.0, 4.0, 3.0, np.nan, 5.0, np.nan, np.nan],
    ]
)
X_expected2 = np.array(
    [
        [1, 4, np.nan],
        [3, 2, np.nan],
    ]
)


@pytest.mark.parametrize("X", [X_incomplete])
def test_utils_prepare_data_2D_succeed(X: NDArray):
    result = utils.prepare_data(X, 1)
    np.testing.assert_allclose(result, X_incomplete)


@pytest.mark.parametrize("X", [X_incomplete])
def test_utils_prepare_data_1D_succeed(X: NDArray):
    X = X.flatten()
    result = utils.prepare_data(X, 5)
    np.testing.assert_allclose(result, X_incomplete)


@pytest.mark.parametrize("X", [X_incomplete])
def test_utils_prepare_data_2D_uneven(X: NDArray):
    result = utils.prepare_data(X, 3)
    np.testing.assert_allclose(result.shape, (15, 2))


@pytest.mark.parametrize("X", [X_incomplete])
def test_utils_prepare_data_consistant(X: NDArray):
    result1 = utils.prepare_data(X, 1)
    result2 = utils.prepare_data(result1, 2)
    result3 = utils.prepare_data(X, 2)
    np.testing.assert_allclose(result2, result3)


@pytest.mark.parametrize("X", [signal])
def test_utils_fold_signal_1D_fail(X: NDArray):
    np.testing.assert_raises(NotDimension2, utils.fold_signal, X, 100)



from typing import Optional, Union
import warnings

import numpy as np
import pandas as pd

from numpy.typing import NDArray

from qolmat.utils.exceptions import NotDimension2, SignalTooShort

HyperValue = Union[int, float, str]


def progress_bar(
    iteration: int,
    total: int,
    prefix: str = "",
    suffix: str = "",
    decimals: int = 1,
    length: int = 100,
    fill: str = "█",
):
    """Call in a loop to create terminal progress bar

    Parameters
    ----------
    iteration : int
        current iteration
    total : int
        total iterations
    prefix : str
        prefix string, by default ""
    suffix : str
        suffix string, by default ""
    decimals : int
        positive number of decimals in percent complete, by default 1
    length : int
        character length of bar, by default 100
    fill : str
        bar fill character, by default "█"
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filled_length = int(length * iteration // total)
    bar = fill * filled_length + "-" * (length - filled_length)
    print(f"\r{prefix} |{bar}| {percent}% {suffix}", end="\r")
    if iteration == total:
        print()


def acf(values: pd.Series, lag_max: int = 30) -> pd.Series:
    """Correlation series of dataseries

    Parameters
    ----------
    values : pd.Series
        dataseries
    lag_max : int, optional
        the maximum lag, by default 30

    Returns
    -------
    pd.Series
        correlation series of value
    """
    acf = pd.Series(0, index=range(lag_max))
    for lag in range(lag_max):
        acf[lag] = values.corr(values.shift(lag))
    return acf


def impute_nans(M: NDArray, method: str = "zeros") -> NDArray:
    """
    Impute the M's nan with the specified method

    Parameters
    ----------
    M : NDArray
        Array to impute
    method : str
        'mean', 'median', or 'zeros'

    Returns
    -------
    NDArray
        Imputed Array
    Raises
    ------
        ValueError
            if ``method`` is not
            in 'mean', 'median' or 'zeros']

    """
    if method == "mean":
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            result = np.where(np.isnan(M), np.resize(np.nanmean(M, axis=0), M.shape), M)
        result = np.where(np.isnan(result), np.nanmean(result), result)
    elif method == "median":
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            result = np.where(np.isnan(M), np.resize(np.nanmedian(M, axis=0), M.shape), M)
        result = np.where(np.isnan(result), np.nanmedian(result), result)
    elif method == "zeros":
        result = np.where(np.isnan(M), 0, M)
    else:
        raise ValueError("'method' should be 'mean', 'median' or 'zeros'.")
    return result


def linear_interpolation(X: NDArray) -> NDArray:
    """
    Impute missing data with a linear interpolation, column-wise

    Parameters
    ----------
    X : NDArray
        array with missing values

    Returns
    -------
    X_interpolated : NDArray
        imputed array, by linear interpolation
    """
    n_rows, n_cols = X.shape
    indices = np.arange(n_cols)
    X_interpolated = X.copy()
    median = np.nanmedian(X)
    for i_row in range(n_rows):
        values = X[i_row]
        mask_isna = np.isnan(values)
        if np.all(mask_isna):
            X_interpolated[i_row] = median
        elif np.any(mask_isna):
            values_interpolated = np.interp(
                indices[mask_isna], indices[~mask_isna], values[~mask_isna]
            )
            X_interpolated[i_row, mask_isna] = values_interpolated
    return X_interpolated


def fold_signal(X: NDArray, period: int) -> NDArray:
    """
    Reshape a time series into a 2D-array

    Parameters
    ----------
    X : NDArray
    period : int
        Period used to fold the signal of the 2D-array

    Returns
    -------
    Tuple[NDArray, int]
        Array and number of added nan's fill it

    Raises
    ------
    ValueError
        if X is not a 1D array
    """
    if len(X.shape) != 2:
        raise NotDimension2(X.shape)
    n_rows, n_cols = X.shape
    n_rows_new = n_rows * period

    X = X.flatten()
    n_required_nans = (-X.size) % n_rows_new
    X = np.append(X, [np.nan] * n_required_nans)
    X = X.reshape(n_rows_new, -1)

    return X


def prepare_data(X: NDArray, period: int = 1) -> NDArray:
    """
    Transform signal to 2D-array in case of 1D-array.
    """
    if len(X.shape) == 1:
        X = X.reshape(1, -1)

    X_fold = fold_signal(X, period)
    return X_fold


def get_shape_original(M: NDArray, shape: tuple) -> NDArray:
    """Shapes an output matrix from the RPCA algorithm into the original shape.

    Parameters
    ----------
    M : NDArray
        Matrix to reshape
    X : NDArray
        Matrix of the desired shape

    Returns
    -------
    NDArray
        Reshaped matrix
    """
    size = np.prod(shape)
    M_flat = M.flatten()[:size]
    return M_flat.reshape(shape)



# ######################
# # Evaluation metrics #
# ######################

import numpy as np
from numpy import random as npr
import pandas as pd
import pytest
import scipy

from qolmat.benchmark import metrics
from qolmat.utils.exceptions import NotEnoughSamples

df_incomplete = pd.DataFrame(
    {"col1": [0, np.nan, 2, 3, np.nan], "col2": [-1, np.nan, 0.5, 1, 1.5]}
)

df_complete = pd.DataFrame({"col1": [0, 2, 2, 3, 4], "col2": [-1, -2, 0.5, 1, 1.5]})

df_imputed = pd.DataFrame({"col1": [0, 1, 2, 3.5, 4], "col2": [-1.5, 0, 1.5, 2, 1.5]})

df_mask = pd.DataFrame(
    {"col1": [False, False, True, True, False], "col2": [True, False, True, True, False]}
)


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_mean_squared_error(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame) -> None:
    assert metrics.mean_squared_error(df1, df1, df_mask).equals(
        pd.Series([0.0, 0.0], index=["col1", "col2"])
    )
    assert metrics.mean_squared_error(df1, df2, df_mask).equals(
        pd.Series([0.125, 0.750], index=["col1", "col2"])
    )


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_root_mean_squared_error(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    assert metrics.root_mean_squared_error(df1, df1, df_mask).equals(
        pd.Series([0.0, 0.0], index=["col1", "col2"])
    )
    assert (
        metrics.root_mean_squared_error(df1, df2, df_mask)
        .round(3)
        .equals(pd.Series([0.354, 0.866], index=["col1", "col2"]))
    )


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_mean_absolute_error(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame) -> None:
    assert metrics.mean_absolute_error(df1, df1, df_mask).equals(
        pd.Series([0.0, 0.0], index=["col1", "col2"])
    )
    assert (
        metrics.mean_absolute_error(df1, df2, df_mask)
        .round(3)
        .equals(pd.Series([0.250, 0.833], index=["col1", "col2"]))
    )


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_mean_absolute_percentage_error(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    assert metrics.mean_absolute_percentage_error(df1, df1, df_mask).equals(
        pd.Series([0.0, 0.0], index=["col1", "col2"])
    )
    result = metrics.mean_absolute_percentage_error(df1, df2, df_mask)
    expected = pd.Series([0.083, 1.166], index=["col1", "col2"])
    np.testing.assert_allclose(result, expected, atol=1e-3)


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_weighted_mean_absolute_percentage_error(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    assert metrics.weighted_mean_absolute_percentage_error(df1, df1, df_mask).equals(
        pd.Series([0.0, 0.0], index=["col1", "col2"])
    )
    result = metrics.weighted_mean_absolute_percentage_error(df1, df2, df_mask)
    expected = pd.Series([0.1, 1.0], index=["col1", "col2"])
    np.testing.assert_allclose(result, expected, atol=1e-3)


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_wasserstein_distance(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame) -> None:
    dist = metrics.dist_wasserstein(df1, df1, df_mask, method="columnwise")
    assert dist.equals(pd.Series([0.0, 0.0], index=["col1", "col2"]))
    dist = metrics.dist_wasserstein(df1, df2, df_mask, method="columnwise")
    assert dist.round(3).equals(pd.Series([0.250, 0.833], index=["col1", "col2"]))


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_kl_divergence_columnwise(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    result = metrics.kl_divergence(df1, df1, df_mask, method="columnwise")
    expected = pd.Series([0.0, 0.0], index=["col1", "col2"])
    np.testing.assert_allclose(result, expected, atol=1e-3)
    result = metrics.kl_divergence(df1, df2, df_mask, method="columnwise")
    expected = pd.Series([18.945, 36.637], index=["col1", "col2"])
    np.testing.assert_allclose(result, expected, atol=1e-3)


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_kl_divergence_gaussian(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    result = metrics.kl_divergence_gaussian(df1, df1, df_mask)
    np.testing.assert_allclose(result, 0, atol=1e-3)

    result = metrics.kl_divergence_gaussian(df1, df2, df_mask)
    np.testing.assert_allclose(result, 1.371, atol=1e-3)


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_kl_divergence_forest(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame) -> None:
    result = metrics.kl_divergence_forest(df1, df1, df_mask)
    np.testing.assert_allclose(result, 0, atol=1e-3)

    result = metrics.kl_divergence_forest(df1, df2, df_mask)
    np.testing.assert_allclose(result, 6.21e-2, rtol=1e-2)


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_frechet_distance(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame) -> None:
    result = metrics.frechet_distance(df1, df1, df_mask)
    np.testing.assert_allclose(result, 0, atol=1e-3)

    result = metrics.frechet_distance(df1, df2, df_mask)
    np.testing.assert_allclose(result, 0.253, atol=1e-3)


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_kolmogorov_smirnov_test(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    result = metrics.kolmogorov_smirnov_test(df1, df1, df_mask)
    expected = pd.Series([0, 0], index=["col1", "col2"])
    np.testing.assert_allclose(result, expected, atol=1e-3)

    result = metrics.kolmogorov_smirnov_test(df1, df2, df_mask)
    expected = pd.Series([0.5, 2 / 3], index=["col1", "col2"])
    np.testing.assert_allclose(result, expected, atol=1e-3)


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_sum_pairwise_distances(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    result = metrics.sum_pairwise_distances(df1, df2, df_mask)
    np.testing.assert_allclose(result, 28, atol=1e-3)


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_sum_energy_distances(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame) -> None:
    sum_distances_df1 = np.sum(
        scipy.spatial.distance.cdist(
            df1[df_mask].fillna(0.0), df1[df_mask].fillna(0.0), metric="cityblock"
        )
    )
    sum_distances_df2 = np.sum(
        scipy.spatial.distance.cdist(
            df2[df_mask].fillna(0.0), df2[df_mask].fillna(0.0), metric="cityblock"
        )
    )
    sum_distances_df1_df2 = np.sum(
        scipy.spatial.distance.cdist(
            df1[df_mask].fillna(0.0), df2[df_mask].fillna(0.0), metric="cityblock"
        )
    )
    energy_distance_scipy = 2 * sum_distances_df1_df2 - sum_distances_df1 - sum_distances_df2
    energy_distance_qolmat = metrics.sum_energy_distances(df1, df2, df_mask)

    assert energy_distance_qolmat.equals(pd.Series(energy_distance_scipy, index=["All"]))


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_mean_difference_correlation_matrix_numerical_features(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    assert metrics.mean_difference_correlation_matrix_numerical_features(df1, df1, df_mask).equals(
        pd.Series([0.0, 0.0], index=["col1", "col2"])
    )
    assert metrics.mean_difference_correlation_matrix_numerical_features(
        df1, df1, df_mask, False
    ).equals(pd.Series([0.0, 0.0], index=["col1", "col2"]))

    assert metrics.mean_difference_correlation_matrix_numerical_features(df1, df2, df_mask).equals(
        pd.Series([0.0, 0.0], index=["col1", "col2"])
    )


df_incomplete_cat = pd.DataFrame(
    {"col1": ["a", np.nan, "a", "b", np.nan], "col2": ["c", np.nan, "d", "b", "d"]}
)

df_imputed_cat = pd.DataFrame(
    {"col1": ["a", "b", "a", "c", "c"], "col2": ["b", "d", "c", "d", "d"]}
)

df_mask_cat = pd.DataFrame(
    {"col1": [False, False, True, True, False], "col2": [True, False, True, True, False]}
)


@pytest.mark.parametrize("df1", [df_incomplete_cat])
@pytest.mark.parametrize("df2", [df_imputed_cat])
@pytest.mark.parametrize("df_mask", [df_mask_cat])
def test_total_variance_distance(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    result = metrics.total_variance_distance(df1, df1, df_mask)
    expected = pd.Series([0.0, 0.0], index=["col1", "col2"])
    np.testing.assert_allclose(result, expected, atol=1e-3)

    result = metrics.total_variance_distance(df1, df2, df_mask)
    expected = pd.Series([1.0, 0], index=["col1", "col2"])
    np.testing.assert_allclose(result, expected, atol=1e-3)


@pytest.mark.parametrize("df1", [df_incomplete_cat])
@pytest.mark.parametrize("df2", [df_imputed_cat])
@pytest.mark.parametrize("df_mask", [df_mask_cat])
def test_mean_difference_correlation_matrix_categorical_features(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    assert metrics.mean_difference_correlation_matrix_categorical_features(
        df1, df1, df_mask
    ).equals(pd.Series([0.0, 0.0], index=["col1", "col2"]))
    assert metrics.mean_difference_correlation_matrix_categorical_features(
        df1, df1, df_mask, False
    ).equals(pd.Series([0.0, 0.0], index=["col1", "col2"]))
    assert metrics.mean_difference_correlation_matrix_categorical_features(
        df1, df2, df_mask
    ).equals(pd.Series([0.0, 0.0], index=["col1", "col2"]))


df_incomplete_cat_num = pd.DataFrame(
    {"col1": ["a", np.nan, "a", "b", np.nan], "col2": [-1, np.nan, 0.5, 1, 1.5]}
)

df_imputed_cat_num = pd.DataFrame(
    {"col1": ["a", "b", "a", "c", "c"], "col2": [-1.5, 0, 1.5, 2, 1.5]}
)

df_mask_cat_num = pd.DataFrame(
    {"col1": [True, False, True, True, False], "col2": [True, False, True, True, False]}
)


@pytest.mark.parametrize("df1", [df_incomplete_cat_num])
@pytest.mark.parametrize("df2", [df_imputed_cat_num])
@pytest.mark.parametrize("df_mask", [df_mask_cat_num])
def test_mean_diff_corr_matrix_categorical_vs_numerical_features(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    assert metrics.mean_diff_corr_matrix_categorical_vs_numerical_features(
        df1, df1, df_mask
    ).equals(pd.Series([0.0], index=["col1"]))
    assert metrics.mean_diff_corr_matrix_categorical_vs_numerical_features(
        df1, df1, df_mask, False
    ).equals(pd.Series([0.0], index=["col1"]))
    assert metrics.mean_diff_corr_matrix_categorical_vs_numerical_features(
        df1, df2, df_mask
    ).equals(pd.Series([0.07009774198932273], index=["col1"]))


df_imputed_bad_shape = pd.DataFrame({"col1": [0, 1, 2, 3.5, 4]})


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed_bad_shape])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_exception_raise_different_shapes(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    with pytest.raises(Exception):
        metrics.mean_difference_correlation_matrix_numerical_features(df1, df2, df_mask)
    with pytest.raises(Exception):
        metrics.frechet_distance(df1, df2, df_mask)


@pytest.mark.parametrize("df1", [df_incomplete_cat])
@pytest.mark.parametrize("df2", [df_imputed_cat])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_exception_raise_no_numerical_column_found(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    with pytest.raises(Exception):
        metrics.kolmogorov_smirnov_test(df1, df2, df_mask)
    with pytest.raises(Exception):
        metrics.mean_difference_correlation_matrix_numerical_features(df1, df2, df_mask)


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_exception_raise_no_categorical_column_found(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    with pytest.raises(Exception):
        metrics.total_variance_distance(df1, df2, df_mask)


df_incomplete_cat_num_bad = pd.DataFrame(
    {"col1": ["a", np.nan, "c", "b", np.nan], "col2": [-1, np.nan, 0.5, 0.5, 1.5]}
)


@pytest.mark.parametrize("df1", [df_incomplete_cat_num])
@pytest.mark.parametrize("df2", [df_incomplete_cat_num_bad])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_value_error_get_correlation_f_oneway_matrix(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    assert metrics.mean_diff_corr_matrix_categorical_vs_numerical_features(
        df1, df2, df_mask
    ).equals(pd.Series([np.nan], index=["col1"]))


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_distance_anticorr(df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame) -> None:
    result = metrics.distance_anticorr(df1, df2, df_mask)
    np.testing.assert_allclose(result, 1.1e-4, rtol=1e-2)


@pytest.mark.parametrize("df1", [df_incomplete])
@pytest.mark.parametrize("df2", [df_imputed])
@pytest.mark.parametrize("df_mask", [df_mask])
def test_pattern_based_weighted_mean_metric(
    df1: pd.DataFrame, df2: pd.DataFrame, df_mask: pd.DataFrame
) -> None:
    with pytest.raises(NotEnoughSamples):
        metrics.pattern_based_weighted_mean_metric(
            df1, df2, df_mask, metric=metrics.distance_anticorr, min_n_rows=5
        )

    expected = pd.Series([1.1e-4], index=["All"])
    result = metrics.pattern_based_weighted_mean_metric(
        df1, df2, df_mask, metric=metrics.distance_anticorr, min_n_rows=1
    )
    np.testing.assert_allclose(result, expected, rtol=1e-2)


rng = npr.default_rng(123)
df_gauss1 = pd.DataFrame(rng.multivariate_normal([0, 0], [[1, 0.2], [0.2, 2]], size=100))
df_gauss2 = pd.DataFrame(rng.multivariate_normal([0, 1], [[1, 0.2], [0.2, 2]], size=100))
df_mask = pd.DataFrame(np.full_like(df_gauss1, True))


def test_pattern_mae_comparison() -> None:
    def fun_mean_mae(df_gauss1, df_gauss2, df_mask) -> float:
        return metrics.mean_squared_error(df_gauss1, df_gauss2, df_mask).mean()

    result1 = fun_mean_mae(df_gauss1, df_gauss2, df_mask)
    result2 = metrics.pattern_based_weighted_mean_metric(
        df_gauss1, df_gauss2, df_mask, metric=fun_mean_mae, min_n_rows=1
    )
    np.testing.assert_allclose(result1, result2, rtol=1e-2)



import numpy as np
from numpy.typing import NDArray
import pytest
from qolmat.imputations.rpca.rpca_utils import (
    approx_rank,
    soft_thresholding,
    svd_thresholding,
    l1_norm,
    toeplitz_matrix,
)
from qolmat.utils.utils import fold_signal

X_incomplete = np.array(
    [
        [1, np.nan, 3, 2, np.nan],
        [7, 2, np.nan, 1, 1],
        [np.nan, 4, 3, np.nan, 5],
        [np.nan, 4, 3, 5, 5],
        [4, 4, 3, np.nan, 5],
    ]
)
X_complete = np.array(
    [[1, 5, 3, 2, 2], [7, 2, 3, 1, 1], [4, 4, 3, 5, 5], [4, 4, 3, 5, 5], [4, 4, 3, 5, 5]]
)


X_exp_row = np.array(
    [[1.0, 0.0, -1.0, 0.0, 0.0], [0.0, 1.0, 0.0, -1.0, 0.0], [0.0, 0.0, 1.0, 0.0, -1.0]]
)

X_exp_column = np.array(
    [[-1.0, 0.0, 0.0], [0.0, -1.0, 0.0], [1.0, 0.0, -1.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
)

threshold = 0.95
T = 2
dimension = 5


@pytest.mark.parametrize("X", [X_complete])
def test_rpca_utils_approx_rank(X: NDArray):
    result = approx_rank(M=X)
    np.testing.assert_allclose(result, 3)


@pytest.mark.parametrize("X", [X_complete])
@pytest.mark.parametrize("threshold", [threshold])
def test_rpca_utils_soft_thresholding(X: NDArray, threshold: float):
    result = soft_thresholding(X=X, threshold=threshold)
    X_expected = np.array(
        [
            [0.05, 4.05, 2.05, 1.05, 1.05],
            [6.05, 1.05, 2.05, 0.05, 0.05],
            [3.05, 3.05, 2.05, 4.05, 4.05],
            [3.05, 3.05, 2.05, 4.05, 4.05],
            [3.05, 3.05, 2.05, 4.05, 4.05],
        ]
    )
    np.testing.assert_allclose(result, X_expected)


@pytest.mark.parametrize("X", [X_complete])
@pytest.mark.parametrize("threshold", [threshold])
def test_rpca_utils_svd_thresholding(X: NDArray, threshold: float):
    result = svd_thresholding(X=X, threshold=threshold)
    X_expected = np.array(
        [
            [1.22954596, 4.19288775, 2.58695225, 2.11783467, 2.11783467],
            [6.14486781, 1.96110299, 2.72428711, 1.21646981, 1.21646981],
            [3.84704901, 3.8901204, 2.92414337, 4.6397143, 4.6397143],
            [3.84704901, 3.8901204, 2.92414337, 4.6397143, 4.6397143],
            [3.84704901, 3.8901204, 2.92414337, 4.6397143, 4.6397143],
        ]
    )
    np.testing.assert_allclose(result, X_expected, rtol=1e-5)


@pytest.mark.parametrize("X", [X_incomplete])
def test_rpca_utils_l1_norm(X: NDArray):
    result = l1_norm(M=X_complete)
    np.testing.assert_allclose(result, 90)


@pytest.mark.parametrize("T", [T])
@pytest.mark.parametrize("dimension", [dimension])
@pytest.mark.parametrize("model, X_expected", [("row", X_exp_row), ("column", X_exp_column)])
def test_rpca_utils_toeplitz_matrix(T: int, dimension: int, model: str, X_expected: NDArray):
    result = toeplitz_matrix(T=T, dimension=dimension, model=model)
    np.testing.assert_allclose(result, X_expected)



from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import pytest

from qolmat.benchmark import hyperparameters
from qolmat.benchmark.hyperparameters import HyperValue

# from hyperparameters import HyperValue
from qolmat.benchmark.missing_patterns import _HoleGenerator, EmpiricalHoleGenerator
from qolmat.imputations.imputers import _Imputer, ImputerRPCA

import hyperopt as ho

df_origin = pd.DataFrame({"col1": [0, np.nan, 2, 4, np.nan], "col2": [-1, np.nan, 0.5, 1, 1.5]})
df_imputed = pd.DataFrame({"col1": [0, 1, 2, 3.5, 4], "col2": [-1.5, 0, 1.5, 2, 1.5]})
df_mask = pd.DataFrame(
    {"col1": [False, False, True, False, False], "col2": [True, False, True, True, False]}
)
df_corrupted = df_origin.copy()
df_corrupted[df_mask] = np.nan

imputer_rpca = ImputerRPCA(tau=2, random_state=42, columnwise=True, period=1)
dict_imputers_rpca = {"rpca": imputer_rpca}
generator_holes = EmpiricalHoleGenerator(n_splits=1, ratio_masked=0.5)
dict_config_opti = {
    "rpca": {
        "lam": {
            "col1": {"min": 0.1, "max": 6, "type": "Real"},
            "col2": {"min": 1, "max": 4, "type": "Real"},
        },
        "tol": {"min": 1e-6, "max": 0.1, "type": "Real"},
        "max_iter": {"min": 99, "max": 100, "type": "Integer"},
        "norm": {"categories": ["L1", "L2"], "type": "Categorical"},
    }
}


class ImputerTest(_Imputer):
    def __init__(
        self,
        groups: Tuple[str, ...] = (),
        random_state: Union[None, int, np.random.RandomState] = None,
        value: float = 0,
    ) -> None:
        super().__init__(groups=groups, columnwise=True, random_state=random_state)
        self.value = value

    def _transform_element(self, df: pd.DataFrame, col: str = "__all__"):
        df_out = df.copy()
        df_out = df_out.fillna(self.value)
        return df_out


class HoleGeneratorTest(_HoleGenerator):
    def __init__(self, mask: pd.Series, subset: Optional[List[str]] = None):
        super().__init__(n_splits=1, subset=subset)
        self.mask = mask

    def generate_mask(self, X: pd.DataFrame) -> pd.DataFrame:
        df_out = X.copy()
        for col in df_out:
            df_out[col] = self.mask
        return df_out


def test_hyperparameters_get_objective() -> None:
    imputer = ImputerTest()
    generator = HoleGeneratorTest(pd.Series([False, False, True, True]), subset=["some_col"])
    metric = "mse"
    names_hyperparams = ["value"]
    df = pd.DataFrame({"some_col": [np.nan, 0, 3, 5]})
    fun_obj = hyperparameters.get_objective(imputer, df, generator, metric, names_hyperparams)
    assert fun_obj([4]) == 1
    assert fun_obj([0]) == (3**2 + 5**2) / 2


def test_hyperparameters_optimize():
    imputer = ImputerTest()
    generator = HoleGeneratorTest(pd.Series([False, False, True, True]), subset=["some_col"])
    metric = "mse"
    dict_config_opti = {"value": ho.hp.uniform("value", 0, 10)}
    df = pd.DataFrame({"some_col": [np.nan, 0, 3, 5]})
    imputer_opti = hyperparameters.optimize(
        imputer, df, generator, metric, dict_config_opti, max_evals=500
    )
    assert isinstance(imputer_opti, ImputerTest)
    np.testing.assert_almost_equal(imputer_opti.value, 4, decimal=1)



import warnings

import numpy as np
import pytest
from numpy.typing import NDArray

from qolmat.imputations.rpca.rpca_pcp import RPCAPCP
from qolmat.utils import utils
from qolmat.utils.data import generate_artificial_ts

X_complete = np.array([[1, 2], [3, 1]], dtype=float)
X_incomplete = np.array([[1, 2], [3, np.nan], [np.nan, 4]], dtype=float)
max_iterations = 50
small_mu = 1e-5
large_mu = 1e5


@pytest.fixture
def synthetic_temporal_data():
    n_samples = 1000
    periods = [100, 20]
    amp_anomalies = 0.5
    ratio_anomalies = 0.05
    amp_noise = 0.1
    X_true, A_true, E_true = generate_artificial_ts(
        n_samples, periods, amp_anomalies, ratio_anomalies, amp_noise
    )
    signal = X_true + A_true + E_true
    mask = np.random.choice(len(signal), round(len(signal) / 20))
    signal[mask] = np.nan
    return signal


@pytest.mark.parametrize(
    "obs, lr, ano, omega, lam",
    [
        (
            np.array([[1, 1], [1, 1]], dtype=float),
            np.array([[2, 2], [2, 2]], dtype=float),
            np.array([[2, 2], [2, 2]], dtype=float),
            np.array([[False, True], [False, False]], dtype=float),
            2,
        )
    ],
)
def test_check_cost_function_minimized_warning(
    obs: NDArray, lr: NDArray, ano: NDArray, omega: NDArray, lam: float
):
    """Test warning when the cost function is minimized."""
    with pytest.warns(UserWarning):
        RPCAPCP()._check_cost_function_minimized(obs, lr, ano, omega, lam)


@pytest.mark.parametrize(
    "obs, lr, ano, omega, lam",
    [
        (
            np.array([[1, 1], [1, 1]], dtype=float),
            np.array([[0, 0], [0, 0]], dtype=float),
            np.array([[2, 2], [2, 2]], dtype=float),
            np.array([[False, False], [False, False]], dtype=float),
            0,
        )
    ],
)
def test_check_cost_function_minimized_no_warning(
    obs: NDArray, lr: NDArray, ano: NDArray, omega: NDArray, lam: float
):
    """Test no warning when the cost function is minimized."""
    with warnings.catch_warnings(record=True) as record:
        RPCAPCP()._check_cost_function_minimized(obs, lr, ano, omega, lam)
    assert len(record) == 0


@pytest.mark.parametrize("X", [X_complete])
def test_rpca_rpca_pcp_get_params_scale(X: NDArray):
    """Test the parameters are well scaled."""
    rpca_pcp = RPCAPCP(max_iterations=max_iterations, mu=0.5, lam=0.1)
    result_dict = rpca_pcp.get_params_scale(X)
    result = list(result_dict.values())
    params_expected = [1 / 7, np.sqrt(2) / 2]
    np.testing.assert_allclose(result, params_expected, atol=1e-4)


@pytest.mark.parametrize("X, mu", [(X_complete, small_mu)])
def test_rpca_rpca_pcp_zero_lambda_small_mu(X: NDArray, mu: float):
    """Test RPCA PCP results if lambda equals zero.
    The problem is ill-conditioned and the result depends
    on the parameter mu; case when mu is small.
    """
    rpca_pcp = RPCAPCP(lam=0, mu=mu)
    X_result, A_result = rpca_pcp.decompose_rpca_signal(X)
    np.testing.assert_allclose(X_result, np.full_like(X, 0), atol=1e-4)
    np.testing.assert_allclose(A_result, X, atol=1e-4)


@pytest.mark.parametrize("X, mu", [(X_complete, large_mu)])
def test_rpca_rpca_pcp_zero_lambda_large_mu(X: NDArray, mu: float):
    """Test RPCA PCP results if lambda equals zero.
    The problem is ill-conditioned and the result depends
    on the parameter mu; case when mu is large.
    """
    rpca_pcp = RPCAPCP(lam=0, mu=mu)
    X_result, A_result = rpca_pcp.decompose_rpca_signal(X)
    np.testing.assert_allclose(X_result, X, atol=1e-4)
    np.testing.assert_allclose(A_result, np.full_like(X, 0), atol=1e-4)


@pytest.mark.parametrize("X, mu", [(X_complete, large_mu)])
def test_rpca_rpca_pcp_large_lambda_small_mu(X: NDArray, mu: float):
    """Test RPCA PCP results with large lambda and small mu."""
    rpca_pcp = RPCAPCP(lam=1e3, mu=mu)
    X_result, A_result = rpca_pcp.decompose_rpca_signal(X)
    np.testing.assert_allclose(X_result, X, atol=1e-4)
    np.testing.assert_allclose(A_result, np.full_like(X, 0), atol=1e-4)


def test_rpca_temporal_signal(synthetic_temporal_data):
    """Test RPCA PCP results for time series data.
    Check if the cost function is smaller at the end than at the start."""
    signal = synthetic_temporal_data
    period = 100
    lam = 0.1
    rpca = RPCAPCP(period=period, lam=lam, mu=0.01)
    X_result, A_result = rpca.decompose_rpca_signal(signal)
    X_input_rpca = utils.linear_interpolation(signal.reshape(period, -1))
    assert np.linalg.norm(X_input_rpca, "nuc") >= np.linalg.norm(
        X_result.reshape(period, -1), "nuc"
    ) + lam * np.sum(np.abs(A_result.reshape(period, -1)))



from __future__ import annotations

import warnings
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import scipy as scp
from numpy.typing import NDArray
from sklearn import utils as sku

from qolmat.imputations.rpca import rpca_utils
from qolmat.imputations.rpca.rpca import RPCA


class RPCANoisy(RPCA):
    """
    This class implements a noisy version of the so-called 'improved RPCA'

    References
    ----------
    Wang, Xuehui, et al. "An improved robust principal component analysis model for anomalies
    detection of subway passenger flow."
    Journal of advanced transportation (2018).

    Chen, Yuxin, et al. "Bridging convex and nonconvex optimization in robust PCA: Noise, outliers
    and missing data."
    The Annals of Statistics 49.5 (2021): 2948-2971.

    Parameters
    ----------
    period: Optional[int]
        number of rows of the reshaped matrix if the signal is a 1D-array
    rank: Optional[int]
        (estimated) low-rank of the matrix D
    tau: Optional[float]
        penalizing parameter for the nuclear norm
    lam: Optional[float]
        penalizing parameter for the sparse matrix
    list_periods: Optional[List[int]]
        list of periods, linked to the Toeplitz matrices
    list_etas: Optional[List[float]]
        list of penalizing parameters for the corresponding period in list_periods
    max_iterations: Optional[int]
        stopping criteria, maximum number of iterations. By default, the value is set to 10_000
    tol: Optional[float]
        stoppign critera, minimum difference between 2 consecutive iterations. By default,
        the value is set to 1e-6
    norm: str
        error norm, can be "L1" or "L2". By default, the value is set to "L2"
    """

    def __init__(
        self,
        random_state: Union[None, int, np.random.RandomState] = None,
        period: int = 1,
        rank: Optional[int] = None,
        mu: Optional[float] = None,
        tau: Optional[float] = None,
        lam: Optional[float] = None,
        list_periods: List[int] = [],
        list_etas: List[float] = [],
        max_iterations: int = int(1e4),
        tol: float = 1e-6,
        norm: str = "L2",
        verbose: bool = True,
    ) -> None:
        super().__init__(period=period, max_iterations=max_iterations, tol=tol, verbose=verbose)
        self.rng = sku.check_random_state(random_state)
        self.rank = rank
        self.mu = mu
        self.tau = tau
        self.lam = lam
        self.list_periods = list_periods
        self.list_etas = list_etas
        self.norm = norm

    def decompose_rpca_L1(
        self, D: NDArray, Omega: NDArray, lam: float, tau: float, rank: int
    ) -> Tuple:
        """
        Compute the noisy RPCA with a L1 time penalisation

        Parameters
        ----------
        D : np.ndarray
            Observations matrix of shape (m, n).
        Omega : np.ndarray
            Binary matrix indicating the observed entries of D, shape (m, n).
        lam : float
            Regularization parameter for the L1 norm.
        tau : float
            Regularization parameter for the temporal correlations.
        rank : int
            Rank parameter for low-rank matrix decomposition.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]
            A tuple containing:
        M : np.ndarray
            Low-rank signal matrix of shape (m, n).
        A : np.ndarray
            Anomalies matrix of shape (m, n).
        U : np.ndarray
            Basis Unitary array of shape (m, rank).
        V : np.ndarray
            Basis Unitary array of shape (n, rank).
        errors : np.ndarray
            Array of iterative errors.

        """
        m, n = D.shape
        rho = 1.1
        mu = self.mu or 1e-2
        mu_bar = mu * 1e3

        # init
        Y = np.ones((m, n))
        Y_ = [np.ones((m, n - period)) for period in self.list_periods]

        X = D.copy()
        A = np.zeros((m, n))
        L = np.ones((m, rank))
        Q = np.ones((n, rank))
        R = [np.ones((m, n - period)) for period in self.list_periods]

        # matrices for temporal correlation
        H = [rpca_utils.toeplitz_matrix(period, n, model="column") for period in self.list_periods]
        HHT = np.zeros((n, n))
        for index, _ in enumerate(self.list_periods):
            HHT += self.list_etas[index] * (H[index] @ H[index].T)

        Ir = np.eye(rank)
        In = np.eye(n)

        for _ in range(self.max_iterations):
            X_temp = X.copy()
            A_temp = A.copy()
            L_temp = L.copy()
            Q_temp = Q.copy()
            R_temp = R.copy()

            sums = np.zeros((m, n))
            for index, _ in enumerate(self.list_periods):
                sums += (mu * R[index] - Y_[index]) @ H[index].T

            X = scp.linalg.solve(
                a=((1 + mu) * In + 2 * HHT).T,
                b=(D - A + mu * L @ Q.T - Y + sums).T,
            ).T

            if np.any(np.isnan(D)):
                A_Omega = rpca_utils.soft_thresholding(D - X, lam)
                A_Omega_C = D - X
                A = np.where(Omega, A_Omega, A_Omega_C)
            else:
                A = rpca_utils.soft_thresholding(D - X, lam)

            L = scp.linalg.solve(
                a=(tau * Ir + mu * (Q.T @ Q)).T,
                b=((mu * X + Y) @ Q).T,
            ).T

            Q = scp.linalg.solve(
                a=(tau * Ir + mu * (L.T @ L)).T,
                b=((mu * X.T + Y.T) @ L).T,
            ).T

            for index, _ in enumerate(self.list_periods):
                R[index] = rpca_utils.soft_thresholding(
                    X @ H[index] - Y_[index] / mu, self.list_etas[index] / mu
                )

            Y += mu * (X - L @ Q.T)
            for index, _ in enumerate(self.list_periods):
                Y_[index] += mu * (X @ H[index] - R[index])

            # update mu
            mu = min(mu * rho, mu_bar)

            # stopping criteria
            Xc = np.linalg.norm(X - X_temp, np.inf)
            Ac = np.linalg.norm(A - A_temp, np.inf)
            Lc = np.linalg.norm(L - L_temp, np.inf)
            Qc = np.linalg.norm(Q - Q_temp, np.inf)
            Rc = -1
            for index, _ in enumerate(self.list_periods):
                Rc = np.maximum(Rc, np.linalg.norm(R[index] - R_temp[index], np.inf))
            tol = np.amax(np.array([Xc, Ac, Lc, Qc, Rc]))

            if tol < self.tol:
                break
        M = X
        U = L
        V = Q
        return M, A, U, V

    def decompose_rpca_L2(
        self, D: NDArray, Omega: NDArray, lam: float, tau: float, rank: int
    ) -> Tuple:
        """
        Compute the noisy RPCA with a L2 time penalisation

        Parameters
        ----------
        D : np.ndarray
            Observations matrix of shape (m, n).
        Omega : np.ndarray
            Binary matrix indicating the observed entries of D, shape (m, n).
        lam : float
            Regularization parameter for the L2 norm.
        tau : float
            Regularization parameter for the temporal correlations.
        rank : int
            Rank parameter for low-rank matrix decomposition.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]
            A tuple containing:
        M : np.ndarray
            Low-rank signal matrix of shape (m, n).
        A : np.ndarray
            Anomalies matrix of shape (m, n).
        U : np.ndarray
            Basis Unitary array of shape (m, rank).
        V : np.ndarray
            Basis Unitary array of shape (n, rank).
        errors : np.ndarray
            Array of iterative errors.

        """
        rho = 1.1
        m, n = D.shape

        # init
        Y = np.full_like(D, 0)
        X = D.copy()
        A = np.full_like(D, 0)
        U, S, Vt = np.linalg.svd(X)
        U = U[:, :rank]
        S = S[:rank]
        Vt = Vt[:rank, :]
        L = U @ np.diag(np.sqrt(S))
        Q = Vt.transpose() @ np.diag(np.sqrt(S))

        mu = self.mu or 1e-2
        mu_bar = mu * 1e3

        # matrices for temporal correlation
        H = [rpca_utils.toeplitz_matrix(period, n, model="column") for period in self.list_periods]
        HHT = np.zeros((n, n))
        for index, _ in enumerate(self.list_periods):
            HHT += self.list_etas[index] * (H[index] @ H[index].T)

        Ir = np.eye(rank)
        In = np.eye(n)

        for _ in range(self.max_iterations):
            # print("Cost function", self.cost_function(D, X, A, Omega, tau, lam))
            X_temp = X.copy()
            A_temp = A.copy()
            L_temp = L.copy()
            Q_temp = Q.copy()

            X = scp.linalg.solve(
                a=((1 + mu) * In + HHT).T,
                b=(D - A + mu * L @ Q.T - Y).T,
            ).T

            if np.any(np.isnan(D)):
                A_Omega = rpca_utils.soft_thresholding(D - X, lam)
                A_Omega_C = D - X
                A = np.where(Omega, A_Omega, A_Omega_C)
            else:
                A = rpca_utils.soft_thresholding(D - X, lam)

            L = scp.linalg.solve(
                a=(tau * Ir + mu * (Q.T @ Q)).T,
                b=((mu * X + Y) @ Q).T,
            ).T

            Q = scp.linalg.solve(
                a=(tau * Ir + mu * (L.T @ L)).T,
                b=((mu * X.T + Y.T) @ L).T,
            ).T

            Y += mu * (X - L @ Q.T)

            mu = min(mu * rho, mu_bar)

            Xc = np.linalg.norm(X - X_temp, np.inf)
            Ac = np.linalg.norm(A - A_temp, np.inf)
            Lc = np.linalg.norm(L - L_temp, np.inf)
            Qc = np.linalg.norm(Q - Q_temp, np.inf)

            tol = max([Xc, Ac, Lc, Qc])

            if tol < self.tol:
                break

        X = L @ Q.T

        M = X
        U = L
        V = Q

        return M, A, U, V

    def get_params_scale(self, D: NDArray) -> Dict[str, float]:
        """
        Get parameters for scaling in RPCA based on the input data.

        Parameters
        ----------
        D : np.ndarray
            Input data matrix of shape (m, n).

        Returns
        -------
        dict
            A dictionary containing the following parameters:
            - "rank" : int
                Rank estimate for low-rank matrix decomposition.
            - "tau" : float
                Regularization parameter for the temporal correlations.
            - "lam" : float
                Regularization parameter for the L1 norm.

        """
        rank = rpca_utils.approx_rank(D)
        tau = 1.0 / np.sqrt(max(D.shape))
        lam = tau
        return {
            "rank": rank,
            "tau": tau,
            "lam": lam,
        }

    def decompose_rpca(self, D: NDArray, Omega: NDArray) -> Tuple[NDArray, NDArray]:
        """
        Compute the noisy RPCA with L1 or L2 time penalisation

        Parameters
        ----------
        X : NDArray
            Matrix of the observations
        Omega: NDArray
            Matrix of missingness, with boolean data

        Returns
        -------
        M: NDArray
            Low-rank signal
        A: NDArray
            Anomalies
        """

        params_scale = self.get_params_scale(D)

        lam = params_scale["lam"] if self.lam is None else self.lam
        rank = params_scale["rank"] if self.rank is None else self.rank
        rank = int(rank)
        tau = params_scale["tau"] if self.tau is None else self.tau

        _, n_columns = D.shape
        for period in self.list_periods:
            if not period < n_columns:
                raise ValueError(
                    "The periods provided in argument in `list_periods` must smaller "
                    f"than the number of columns in the matrix but {period} >= {n_columns}!"
                )

        if self.norm == "L1":
            M, A, U, V = self.decompose_rpca_L1(D, Omega, lam, tau, rank)

        elif self.norm == "L2":
            M, A, U, V = self.decompose_rpca_L2(D, Omega, lam, tau, rank)

        self._check_cost_function_minimized(D, M, A, Omega, tau, lam)

        return M, A

    def cost_function(
        self,
        observations: NDArray,
        low_rank: NDArray,
        anomalies: NDArray,
        Omega: NDArray,
        tau: float,
        lam: float,
    ):
        temporal_norm: float = 0
        if len(self.list_etas) > 0:
            # matrices for temporal correlation
            H = [
                rpca_utils.toeplitz_matrix(period, observations.shape[1], model="column")
                for period in self.list_periods
            ]
            if self.norm == "L1":
                for eta, H_matrix in zip(self.list_etas, H):
                    temporal_norm += eta * np.sum(np.abs(H_matrix @ low_rank))
            elif self.norm == "L2":
                for eta, H_matrix in zip(self.list_etas, H):
                    temporal_norm += eta * float(np.linalg.norm(low_rank @ H_matrix, "fro"))
        anomalies_norm = np.sum(np.abs(anomalies * Omega))
        cost = (
            1 / 2 * ((Omega * (observations - low_rank - anomalies)) ** 2).sum()
            + tau * np.linalg.norm(low_rank, "nuc")
            + lam * anomalies_norm
            + temporal_norm
        )
        return cost

    def _check_cost_function_minimized(
        self,
        observations: NDArray,
        low_rank: NDArray,
        anomalies: NDArray,
        Omega: NDArray,
        tau: float,
        lam: float,
    ):
        """Check that the functional minimized by the RPCA
        is smaller at the end than at the beginning

        Parameters
        ----------
        observations : NDArray
            observations matrix with first linear interpolation
        low_rank : NDArray
            low_rank matrix resulting from RPCA
        anomalies : NDArray
            sparse matrix resulting from RPCA
        Omega: NDArrau
            boolean matrix indicating the observed values
        tau : float
            parameter penalizing the nuclear norm of the low rank part
        lam : float
            parameter penalizing the L1-norm of the anomaly/sparse part
        norm : str
            norm of the temporal penalisation. Has to be `L1` or `L2`
        """
        cost_start = self.cost_function(
            observations, observations, np.full_like(observations, 0), Omega, tau, lam
        )
        cost_end = self.cost_function(observations, low_rank, anomalies, Omega, tau, lam)
        function_str = "1/2 $ ||D-M-A||_2 + tau ||D||_* + lam ||A||_1"
        if len(self.list_etas) > 0:
            for eta in self.list_etas:
                function_str += f"{eta} ||XH||_{self.norm}"

        if self.verbose and (round(cost_start, 4) - round(cost_end, 4)) <= -1e-2:
            warnings.warn(
                f"RPCA algorithm may provide bad results. Function {function_str} increased from"
                f" {cost_start} to {cost_end} instead of decreasing!".format("%.2f")
            )



from typing import Any, List, Tuple
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
import scipy.sparse
from qolmat.utils import plot
from pytest_mock.plugin import MockerFixture

plt.switch_backend("Agg")

np.random.seed(42)
matrice1 = np.random.rand(3, 3)
matrice2 = np.random.rand(3, 3)
matrice3 = np.random.rand(3, 3)
list_matrices = [matrice1, matrice2, matrice3]

signal1 = [1, -1, 1, 0, 1, 1 / 2, -1 / 2, 0, 1 / 2]
signal2 = [0, -1 / 2, 1 / 2, 1, -1, 1 / 2, 1 / 2, 1, -1 / 2]
list_signals = [signal1, signal2]

X = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

M = np.random.rand(100, 100)
A = np.random.rand(100, 100)
E = scipy.sparse.csr_matrix(np.random.rand(100, 10))

df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6], "C": [7, 8, 9]})
df1 = pd.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6]})
df2 = pd.DataFrame({"x": [2, 3, 4], "y": [5, 6, 7]})
dict_df_imputed = {
    "Imputer1": pd.DataFrame({"A": [2, 3, np.nan], "B": [5, np.nan, 7], "C": [np.nan, 8, 9]})
}


@pytest.mark.parametrize("list_matrices", [list_matrices])
def test_utils_plot_plot_matrices(list_matrices: List[np.ndarray], mocker: MockerFixture) -> None:
    mocker.patch("matplotlib.pyplot.savefig")
    mocker.patch("matplotlib.pyplot.show")
    plot.plot_matrices(list_matrices=list_matrices, title="title")
    assert len(plt.gcf().get_axes()) > 0
    assert plt.savefig.call_count == 1
    plt.close("all")


@pytest.mark.parametrize("list_signals", [list_signals])
def test_utils_plot_plot_signal(list_signals: List[List[Any]], mocker: MockerFixture) -> None:
    mocker.patch("matplotlib.pyplot.savefig")
    mocker.patch("matplotlib.pyplot.show")
    plot.plot_signal(list_signals=list_signals, ylabel="ylabel", title="title")
    assert len(plt.gcf().get_axes()) > 0
    assert plt.savefig.call_count == 1
    plt.close("all")


@pytest.mark.parametrize("M, A, E, index_array, dims", [(M, A, E, [0, 1, 2], (10, 10))])
def test__utils_plot_plot_images(
    M: np.ndarray,
    A: np.ndarray,
    E: np.ndarray,
    index_array: List[int],
    dims: Tuple[int, int],
    mocker: MockerFixture,
):
    mocker.patch("matplotlib.pyplot.savefig")
    mocker.patch("matplotlib.pyplot.show")
    plot.plot_images(M, A, E, index_array, dims, filename="filename")
    assert len(plt.gcf().get_axes()) > 0
    assert plt.savefig.call_count == 1
    plt.close("all")


@pytest.mark.parametrize("X", [X])
def test_utils_plot_make_ellipses(X: np.ndarray, mocker: MockerFixture):
    mocker.patch("matplotlib.pyplot.show")
    ax = plt.gca()
    plot.make_ellipses(X, ax, color="blue")
    assert len(plt.gcf().get_axes()) > 0
    plt.close("all")


@pytest.mark.parametrize("df1,df2", [(df1, df2)])
def test_utils_plot_compare_covariances(
    df1: pd.DataFrame, df2: pd.DataFrame, mocker: MockerFixture
):
    mocker.patch("matplotlib.pyplot.show")
    ax = plt.gca()
    plot.compare_covariances(df1, df2, "x", "y", ax)
    assert len(plt.gcf().get_axes()) > 0
    plt.close("all")


@pytest.mark.parametrize("df", [df])
@pytest.mark.parametrize("orientation", ["horizontal", "vertical"])
def test_utils_plot_multibar(df: pd.DataFrame, orientation: str, mocker: MockerFixture):
    mocker.patch("matplotlib.pyplot.show")
    plot.multibar(df, orientation=orientation)
    assert len(plt.gcf().get_axes()) > 0
    plt.close("all")


@pytest.mark.parametrize("df", [df])
def test_utils_plot_plot_imputations(df: pd.DataFrame, mocker: MockerFixture):
    mocker.patch("matplotlib.pyplot.show")
    plot.plot_imputations(df, dict_df_imputed)
    assert len(plt.gcf().get_axes()) > 0
    plt.close("all")
